package cz.zcu.connectors.moodle;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.identityconnectors.common.logging.Log;
import org.identityconnectors.framework.common.exceptions.ConnectorException;
import org.identityconnectors.framework.common.exceptions.InvalidAttributeValueException;
import org.identityconnectors.framework.common.exceptions.UnknownUidException;
import org.identityconnectors.framework.common.objects.Attribute;
import org.identityconnectors.framework.common.objects.AttributeDelta;
import org.identityconnectors.framework.common.objects.AttributeDeltaBuilder;
import org.identityconnectors.framework.common.objects.AttributeInfo;
import org.identityconnectors.framework.common.objects.AttributeInfoBuilder;
import org.identityconnectors.framework.common.objects.ConnectorObject;
import org.identityconnectors.framework.common.objects.Name;
import org.identityconnectors.framework.common.objects.ObjectClass;
import org.identityconnectors.framework.common.objects.ObjectClassInfo;
import org.identityconnectors.framework.common.objects.ObjectClassInfoBuilder;
import org.identityconnectors.framework.common.objects.OperationOptions;
import org.identityconnectors.framework.common.objects.ResultsHandler;
import org.identityconnectors.framework.common.objects.SyncDeltaBuilder;
import org.identityconnectors.framework.common.objects.SyncDeltaType;
import org.identityconnectors.framework.common.objects.SyncResultsHandler;
import org.identityconnectors.framework.common.objects.SyncToken;
import org.identityconnectors.framework.common.objects.Uid;

import cz.zcu.connectors.Db;
import cz.zcu.connectors.DbFilter;
import cz.zcu.connectors.moodle.dao.Cohort;


public class CohortHandler {
	/** Logging. */
	private static final Log l = Log.getLog(CohortHandler.class);

	/** Database table name. */
	public static final String TABLE = "mdl_cohort";

	// private ObjectClass objectClass;
	/** Connector connection. */
	private final MoodleConnection connection;
	// private final MoodleConfiguration config;
	/** Database connection. */
	private final Db db;
	/** Table name according to the current configuration. */
	private final String table;


	/**
	 * Cohort processing class constructor.
	 *
	 * @param objectClass ConnId object class
	 * @param config connector configuration
	 * @param connection connector connection
	 */
	public CohortHandler(final ObjectClass objectClass, final MoodleConfiguration config, final MoodleConnection connection) {
		// this.objectClass = objectClass;
		this.connection = connection;
		// this.config = config;
		this.db = connection.getDb();
		if (config.getTablePrefix().isEmpty()) {
			table = TABLE;
		} else {
			table = config.getTablePrefix() + "." + TABLE;
		}
	}


	/**
	 * Build schema for Cohort object.
	 *
	 * @return Cohort object schema
	 */
	public final ObjectClassInfo buildSchema() {
		Set<AttributeInfo> attributes;

		attributes = new HashSet<>();
		attributes.add(new AttributeInfoBuilder(Uid.NAME)
				.setNativeName(Cohort.ATTR_ID)
				.setCreateable(false)
				.setRequired(false)
				.setUpdateable(false)
				.build());
		attributes.add(new AttributeInfoBuilder(Name.NAME)
				.setNativeName(Cohort.ATTR_NAME)
				.setRequired(true)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_CONTEXTID, Long.class)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_IDNUMBER)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_DESCRIPTION)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_DESCRIPTIONFORMAT, Integer.class)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_VISIBLE, Integer.class)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_COMPONENT)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_THEME)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_TIMECREATED, ZonedDateTime.class)
				.setCreateable(false)
				.setUpdateable(false)
				.build());
		attributes.add(new AttributeInfoBuilder(Cohort.ATTR_TIMEMODIFIED, ZonedDateTime.class)
				.setCreateable(false)
				.setUpdateable(false)
				.build());

		return new ObjectClassInfoBuilder()
				.setType(Cohort.OBJECT_CLASS_NAME)
				.addAllAttributeInfo(attributes)
				.build();
	}


	/**
	 * Bind parameters from the filter and execute prepared query.
	 *
	 * @param sql SQL command from sqlQuery()
	 * @param st prepared statement from the SQL command
	 * @param filter ConnId filter
	 * @return result set
	 * @throws SQLException SQL Exception
	 */
	private ResultSet dbQuery(final String sql, final PreparedStatement st, final DbFilter filter) throws SQLException {
		if (filter == null) {
			l.ok("SQL: {0}", sql);
		} else if (filter.isValid()) {
			int i = 1;
			List<String> sl = new ArrayList<>();
			for (Object right : filter.right()) {
				if (right == null) {
					throw new InvalidAttributeValueException("Missing value in query filter");
				}
				String v = right.toString();
				st.setString(i++, v);
				sl.add(v);
			}
			l.ok("SQL: {0} [{1}]", sql, String.join(", ", sl));
		} else {
			throw new InvalidAttributeValueException("Empty filter");
		}

		return st.executeQuery();
	}


	/**
	 * Return SQL SELECT command based on the filter.
	 *
	 * @param filter ConnId filter
	 * @return SQL string
	 */
	private String sqlQuery(final DbFilter filter) {
		final String baseSql = "SELECT "
			+ Cohort.FIELD_ID + ", "
			+ Cohort.FIELD_NAME + ", "
			+ Cohort.ATTR_CONTEXTID + ", "
			+ Cohort.ATTR_IDNUMBER + ", "
			+ Cohort.ATTR_DESCRIPTION + ", "
			+ Cohort.ATTR_DESCRIPTIONFORMAT + ", "
			+ Cohort.ATTR_VISIBLE + ", "
			+ Cohort.ATTR_COMPONENT + ", "
			+ Cohort.ATTR_THEME + ", "
			+ Cohort.ATTR_TIMECREATED + ", "
			+ Cohort.ATTR_TIMEMODIFIED + " "
			+ "FROM " + table;

		if (filter == null) {
			return baseSql;
		} else if (filter.isValid()) {
			return baseSql + " WHERE " + filter.where();
		} else {
			throw new InvalidAttributeValueException("Empty query filter");
		}
	}


	/**
	 * Launch and handle database query.
	 *
	 * @param filter database filter object
	 * @param handler ConnId handler
	 * @param options ConnId operation options
	 */
	public void query(final DbFilter filter, final ResultsHandler handler, final OperationOptions options) throws SQLException {
		l.info("query(query {0}, options {1})", filter, options);
		final String sql = sqlQuery(filter);
		final PreparedStatement st = db.create(sql);

		ResultSet rs = dbQuery(sql, st, filter);
		while (rs.next()) {
			Cohort cohort = new Cohort();
			cohort.load(rs, connection.getZone());
			l.ok("fetched cohort {0} ({1})", cohort.getUid(), cohort.getName());
			final ConnectorObject co = cohort.createConnectorObject();
			if (!handler.handle(co)) {
				l.info("listing stopped at {0} ({1})", cohort.getUid(), cohort.getName());
				break;
			}
		}
		rs.close();
	}


	private void checkDuplicity(final String uid, final String column, final String value) throws SQLException {
		Utils.checkDuplicity("cohort", db, table, uid, Cohort.FIELD_ID, column, value);
	}


	private int dbUpdate(final PreparedStatement st, final String sql, final Cohort cohort, final String oldUid) throws SQLException {
		st.setString(1, cohort.getUid());
		st.setString(2, cohort.getName());
		if (cohort.get(Cohort.ATTR_CONTEXTID) == null) {
			st.setNull(3, java.sql.Types.BIGINT);
		} else {
			st.setLong(3, cohort.get(Cohort.ATTR_CONTEXTID));
		}
		st.setString(4, cohort.get(Cohort.ATTR_IDNUMBER));
		st.setString(5, cohort.get(Cohort.ATTR_DESCRIPTION));
		if (cohort.get(Cohort.ATTR_DESCRIPTIONFORMAT) == null) {
			st.setNull(6, java.sql.Types.INTEGER);
		} else {
			st.setInt(6, cohort.get(Cohort.ATTR_DESCRIPTIONFORMAT));
		}
		if (cohort.get(Cohort.ATTR_VISIBLE) == null) {
			st.setNull(7, java.sql.Types.INTEGER);
		} else {
			st.setInt(7, cohort.get(Cohort.ATTR_VISIBLE));
		}
		st.setString(8, cohort.get(Cohort.ATTR_COMPONENT));
		st.setString(9, cohort.get(Cohort.ATTR_THEME));
		if (oldUid == null) {
			l.ok("SQL: {0} [{1}, ...]", sql, String.join(", ", cohort.getUid(), cohort.getName()));
		} else {
			st.setString(10, oldUid);
			l.ok("SQL: {0} [{1}, ..., {2}]", sql, String.join(", ", cohort.getUid()), oldUid);
		}
		st.execute();

		if (cohort.getUid() == null) {
			String uid = Utils.lastId(db, st, Cohort.FIELD_ID, table);
			cohort.set(Cohort.ATTR_ID, uid);
			cohort.setUid(uid);
		}

		return st.getUpdateCount();
	}


	/**
	 * Create Cohort object on resource.
	 *
	 * @param attributes attribute set
	 * @param options ConnId operation options
	 * @return UID of the new object
	 */
	public Uid create(final Set<Attribute> attributes, final OperationOptions options) throws SQLException {
		final String sql = "INSERT INTO " + table
			+ " ("
				+ Cohort.FIELD_ID + ", "
				+ Cohort.FIELD_NAME + ", "
				+ Cohort.ATTR_CONTEXTID + ", "
				+ Cohort.ATTR_IDNUMBER + ", "
				+ Cohort.ATTR_DESCRIPTION + ", "
				+ Cohort.ATTR_DESCRIPTIONFORMAT + ", "
				+ Cohort.ATTR_VISIBLE + ", "
				+ Cohort.ATTR_COMPONENT + ", "
				+ Cohort.ATTR_THEME + ", "
				+ Cohort.ATTR_TIMECREATED + ", "
				+ Cohort.ATTR_TIMEMODIFIED
			+ ") VALUES ("
				+ "?, ?, ?, ?, ?, ?, ?, ?, ?, UNIX_TIMESTAMP(), UNIX_TIMESTAMP()"
			+ ")";
		if (attributes == null || attributes.isEmpty()) {
			l.warn("empty attributes, create request ignored");
			return null;
		}
		Cohort cohort = new Cohort();
		cohort.setAttributes(attributes);
		cohort.validate();

		checkDuplicity(null, Cohort.FIELD_ID, cohort.getUid());
		checkDuplicity(null, Cohort.FIELD_NAME, cohort.getName());

		PreparedStatement st = db.create("insert cohort", sql, Statement.RETURN_GENERATED_KEYS);
		int count = dbUpdate(st, sql, cohort, null);
		l.ok("SQL: inserted {1} rows", count);

		return new Uid(cohort.getUid());
	}


	/**
	 * Set explicitly NULL on the database field on identity.
	 *
	 * @param key database field name
	 * @param uid UID of the identity
	 * @return number of updated rows
	 * @throws SQLException SQL Exception
	 */
	private int dbSetNull(final String key, final String uid) throws SQLException {
		final String sqlNull = "UPDATE " + table + " SET " + key + " = NULL WHERE " + Cohort.FIELD_ID + " = ?";
		PreparedStatement st = db.create(sqlNull);
		st.setString(1, uid);
		l.ok("SQL: {0} [{1}]", sqlNull, uid);
		st.execute();
		return st.getUpdateCount();
	}


	/**
	 * Update Cohort object on the resource.
	 *
	 * @param uid UID
	 * @param modifications attribute delta set with modifications
	 * @param options ConnId operation options
	 * @return transitively changed attributes
	 */
	public Set<AttributeDelta> updateDelta(final Uid uid, final Set<AttributeDelta> modifications, final OperationOptions options) throws SQLException {
		final String sql = "UPDATE " + table + " SET "
			+ Cohort.FIELD_ID + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.FIELD_ID + "), "
			+ Cohort.FIELD_NAME + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.FIELD_NAME + "), "
			+ Cohort.ATTR_CONTEXTID + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.ATTR_CONTEXTID + "), "
			+ Cohort.ATTR_IDNUMBER + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.ATTR_IDNUMBER + "), "
			+ Cohort.ATTR_DESCRIPTION + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.ATTR_DESCRIPTION + "), "
			+ Cohort.ATTR_DESCRIPTIONFORMAT + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.ATTR_DESCRIPTIONFORMAT + "), "
			+ Cohort.ATTR_VISIBLE + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.ATTR_VISIBLE + "), "
			+ Cohort.ATTR_COMPONENT + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.ATTR_COMPONENT + "), "
			+ Cohort.ATTR_THEME + " = " + db.getSql(Db.SQL.IF_NULL) + "(?, " + Cohort.ATTR_THEME + "), "
			+ Cohort.ATTR_TIMEMODIFIED + " = UNIX_TIMESTAMP() "
			+ "WHERE " + Cohort.FIELD_ID + " = ?";
		Set<AttributeDelta> changes = null;
		int count = 0;

		if (uid == null) {
			throw new InvalidAttributeValueException("UID is null");
		}
		String oldUid = uid.getUidValue();

		if (modifications == null || modifications.isEmpty()) {
			l.warn("empty attributes, updateDelta request ignored");
			return null;
		}
		Cohort cohort = new Cohort();
		cohort.setDelta(modifications);

		if (cohort.getUid() != null && (oldUid == null || !oldUid.equals(cohort.getUid()))) {
			checkDuplicity(oldUid, Cohort.FIELD_ID, cohort.getUid());
			changes = new HashSet<>();
			changes.add(AttributeDeltaBuilder.build(Uid.NAME, cohort.getUid()));
		}
		// check before rename
		if (cohort.getName() != null) {
			checkDuplicity(oldUid, Cohort.ATTR_NAME, cohort.getName());
		}

		// update
		PreparedStatement st = db.create("update cohort", sql, Statement.RETURN_GENERATED_KEYS);
		count += dbUpdate(st, sql, cohort, oldUid);
		if (count == 0) {
			throw new UnknownUidException("Cohort " + cohort.getUid() + " to update doesn't exist");
		}
		if (cohort.getUid() == null) {
			cohort.setUid(oldUid);
		}

		// explicit NULL needed
		for (String key : Cohort.ATTRS.keySet()) {
			if (cohort.contains(key) && cohort.get(key) == null) {
				count += dbSetNull(key, cohort.getUid());
			}
		}

		l.info("{0} updates", count);
		return changes;
	}


	/**
	 * Delete Cohort object from the resource.
	 *
	 * @param uid UID
	 * @param options ConnId operation options
	 * @return number of rows deleted
	 */
	public int delete(final Uid uid, final OperationOptions options) throws SQLException {
		final String sql = "DELETE FROM " + table + " WHERE " + Cohort.FIELD_ID + " = ?";

		if (uid == null) {
			throw new InvalidAttributeValueException("UID is null");
		}
		String uidValue = uid.getUidValue();

		PreparedStatement st = db.create(sql);
		st.setString(1, uidValue);
		l.ok("SQL: {0} [{1}]", sql, uidValue);
		st.execute();
		int count = st.getUpdateCount();
		l.ok("SQL delete count {0}", count);

		if (count == 0) {
			throw new UnknownUidException("Cohort " + uidValue + " to delete doesn't exist");
		}

		return count;
	}


	/**
	 * Get latest synchronization token.
	 *
	 * @return synchronization token
	 */
	public SyncToken getLatestSyncToken() throws SQLException {
		final String sql = "SELECT MAX(" + Cohort.ATTR_TIMEMODIFIED + ") FROM " + table + " WHERE " + Cohort.ATTR_TIMEMODIFIED + " < " + connection.getSqlNow();
		Long t = null;
		long epoch;

		PreparedStatement st = db.create("get sync cohort", sql);
		l.ok("SQL: {0}", sql);
		ResultSet rs = st.executeQuery();
		if (rs.next()) {
			t = rs.getLong(1);
		}
		rs.close();

		// null token value is forbidden in ConnId
		// ==> get the current time from the database
		if (t == null) {
			final String sqlSelectNow;

			switch (db.getDbType()) {
			case Db.ORACLE:
				sqlSelectNow = "SELECT " + connection.getSqlNow() + " FROM dual";
				break;
			default:
				sqlSelectNow = "SELECT " + connection.getSqlNow();
			}
			st = db.create(sqlSelectNow);
			l.ok("SQL: {0}", sqlSelectNow);
			rs = st.executeQuery();
			if (!rs.next()) {
				throw new ConnectorException("No result from DB time query");
			}
			t = rs.getLong(1);
			rs.close();
			// consider also all upcoming changes from the current millisecond
			epoch = t - 1;
		} else {
			// do not consider any current changes from DB
			epoch = t;
		}
		l.info("latest sync {0}", epoch);

		return new SyncToken(epoch);
	}


	/**
	 * Get Cohort objects changed on the resource.
	 *
	 * @param token synchronization token
	 * @param handler ConnId sync result handler
	 * @param options ConnId operation options
	 */
	public void sync(final SyncToken token, final SyncResultsHandler handler, final OperationOptions options) throws SQLException {
		final String sql = sqlQuery(null) + " WHERE " + Cohort.ATTR_TIMEMODIFIED + " > ? AND " + Cohort.ATTR_TIMEMODIFIED + " < " + connection.getSqlNow() + " ORDER BY timemodified";
		Long t = null;

		if (token != null) {
			t = (Long) token.getValue();
		}
		PreparedStatement st = db.create("sync cohort", sql);
		if (t == null) {
			st.setNull(1, java.sql.Types.BIGINT);
		} else {
			st.setLong(1, t);
		}
		l.ok("SQL: {0} [{1}]", sql, t);

		//{
		//	PreparedStatement stNow = db.create("SELECT " + connection.getSqlNow());
		//	ResultSet rsNow = stNow.executeQuery();
		//	rsNow.next();
		//	java.sql.Timestamp t2 = rsNow.getLong(1);
		//	l.ok("SQL: {0} is {1}", connection.getSqlNow(), t2 == null ? null : t2.getTime());
		//}
		ResultSet rs = st.executeQuery();
		while (rs.next()) {
			Cohort cohort = new Cohort();
			cohort.load(rs, connection.getZone());
			l.ok("fetched cohort {0} ({1})", cohort.getUid(), cohort.getName());
			final ConnectorObject co = cohort.createConnectorObject();
			final SyncToken newToken = new SyncToken(cohort.getAsTime(Cohort.ATTR_TIMEMODIFIED) / 1000);

			final SyncDeltaBuilder deltaBuilder = new SyncDeltaBuilder();
			deltaBuilder
				.setToken(newToken)
				.setDeltaType(SyncDeltaType.CREATE_OR_UPDATE)
				.setObject(co);
			if (handler.handle(deltaBuilder.build())) {
				l.ok("syncing cohort {0} ({1})", cohort.getUid(), cohort.getName());
			} else {
				l.info("syncing cohort stopped at {0} ({1})", cohort.getUid(), cohort.getName());
				break;
			}
		}
		rs.close();
	}

	/**
	 * Test connection to the database resource.
	 */
	public void test() throws SQLException {
		final String sql = "SELECT id FROM " + table;
		PreparedStatement st = db.create("test cohort", sql);
		l.ok("SQL: {0}", sql);
		ResultSet rs = st.executeQuery();
		rs.next();
		rs.close();
	}

}
