package cz.zcu.connectors.moodle;

import java.sql.SQLException;
import java.util.Set;

import org.identityconnectors.framework.common.exceptions.ConnectorException;
import org.identityconnectors.framework.common.objects.Attribute;
import org.identityconnectors.framework.common.objects.AttributeDelta;
import org.identityconnectors.framework.common.objects.ObjectClass;
import org.identityconnectors.framework.common.objects.OperationOptions;
import org.identityconnectors.framework.common.objects.ResultsHandler;
import org.identityconnectors.framework.common.objects.Schema;
import org.identityconnectors.framework.common.objects.SchemaBuilder;
import org.identityconnectors.framework.common.objects.SyncResultsHandler;
import org.identityconnectors.framework.common.objects.SyncToken;
import org.identityconnectors.framework.common.objects.Uid;
import org.identityconnectors.framework.common.objects.filter.FilterTranslator;
import org.identityconnectors.framework.spi.Configuration;
import org.identityconnectors.framework.spi.ConnectorClass;
import org.identityconnectors.framework.spi.PoolableConnector;
import org.identityconnectors.framework.spi.operations.CreateOp;
import org.identityconnectors.framework.spi.operations.DeleteOp;
import org.identityconnectors.framework.spi.operations.SchemaOp;
import org.identityconnectors.framework.spi.operations.SearchOp;
import org.identityconnectors.framework.spi.operations.SyncOp;
import org.identityconnectors.framework.spi.operations.TestOp;
import org.identityconnectors.framework.spi.operations.UpdateDeltaOp;

import cz.zcu.connectors.Db;
import cz.zcu.connectors.DbFilter;
import cz.zcu.connectors.moodle.dao.Cohort;
import cz.zcu.connectors.moodle.dao.User;


@ConnectorClass(displayNameKey = "moodle.connector.display", configurationClass = MoodleConfiguration.class)
public class MoodleConnector implements PoolableConnector, CreateOp, DeleteOp, UpdateDeltaOp, SchemaOp, SearchOp<DbFilter>, SyncOp, TestOp {
	/** Connector configuration. */
	private MoodleConfiguration configuration;
	/** Connector connection. */
	private MoodleConnection connection;
	/** Connector schema. */
	private Schema schema;


	/**
	 * Initialize connector.
	 *
	 * @param genericConfiguration connector configuration
	 */
	@Override
	public void init(final Configuration genericConfiguration) {
		final MoodleConfiguration config = (MoodleConfiguration) genericConfiguration;
		this.configuration = config;
		this.connection = new MoodleConnection(config);
	}


	/**
	 * Dispose connector.
	 */
	@Override
	public void dispose() {
		if (connection != null) {
			connection.close();
			connection = null;
		}
		configuration = null;
		//schema = null; // re-fetch schema
	}


	/**
	 * Get connector configuration.
	 *
	 * @return connector configuration
	 */
	@Override
	public Configuration getConfiguration() {
		return configuration;
	}


	/**
	 * Connection alive check.
	 */
	@Override
	public void checkAlive() {
		if (configuration == null || connection == null) {
			throw new ConnectorException("checkAlive(): connector not initialized");
		}
		connection.checkAlive();
	}


	/**
	 * Construct and return connector schema.
	 *
	 * @return connector schema
	 */
	@Override
	public Schema schema() {
		if (schema == null) {
			final SchemaBuilder schemaBuilder = new SchemaBuilder(MoodleConnector.class);

			final UserHandler uh = new UserHandler(User.OBJECT_CLASS, configuration, connection);
			final CohortHandler ch = new CohortHandler(Cohort.OBJECT_CLASS, configuration, connection);
			schemaBuilder.defineObjectClass(uh.buildSchema());
			schemaBuilder.defineObjectClass(ch.buildSchema());

			schema = schemaBuilder.build();
		}
		return schema;
	}


	/**
	 * Construct and return filter translator.
	 *
	 * @param objectClass ConnId object class
	 * @param options ConnId operation options
	 * @return connector filter translator
	 */
	@Override
	public FilterTranslator<DbFilter> createFilterTranslator(final ObjectClass objectClass, final OperationOptions options) {
		if (objectClass.is(User.OBJECT_CLASS_NAME)) {
			return new UserFilterTranslator();
		} else if (objectClass.is(Cohort.OBJECT_CLASS_NAME)) {
			return new CohortFilterTranslator();
		} else {
			throw new UnsupportedOperationException("Unsupported object class " + objectClass);
		}
	}


	/**
	 * Create object on resource.
	 *
	 * @param objectClass object class
	 * @param attributes attributes set
	 * @param options operation options
	 *
	 * @return UID of the new connector object
	 */
	@Override
	public Uid create(final ObjectClass objectClass, final Set<Attribute> attributes, final OperationOptions options) {
		Uid uid = null;
		try {
			if (objectClass.is(User.OBJECT_CLASS_NAME)) {
				final UserHandler h = new UserHandler(objectClass, configuration, connection);
				uid = h.create(attributes, options);
			} else if (objectClass.is(Cohort.OBJECT_CLASS_NAME)) {
				final CohortHandler h = new CohortHandler(objectClass, configuration, connection);
				uid = h.create(attributes, options);
			} else {
				throw new UnsupportedOperationException("Unsupported object class " + objectClass);
			}
		} catch (SQLException ex) {
			Db.handleException(ex, "Insert error");
		}
		return uid;
	}


	/**
	 * Query resource objects.
	 *
	 * @param objectClass object class
	 * @param query filter query
	 * @param handler results handler
	 * @param options operation options
	 */
	@Override
	public void executeQuery(final ObjectClass objectClass, final DbFilter query, final ResultsHandler handler, final OperationOptions options) {
		try {
			if (objectClass.is(User.OBJECT_CLASS_NAME)) {
				final UserHandler h = new UserHandler(objectClass, configuration, connection);
				h.query(query, handler, options);
			} else if (objectClass.is(Cohort.OBJECT_CLASS_NAME)) {
				final CohortHandler h = new CohortHandler(objectClass, configuration, connection);
				h.query(query, handler, options);
			} else {
				throw new UnsupportedOperationException("Unsupported object class " + objectClass);
			}
		} catch (SQLException ex) {
			Db.handleException(ex, "Query error");
		}
	}


	/**
	 * Test resource connection.
	 */
	@Override
	public void test() {
		try {
			final UserHandler uh = new UserHandler(User.OBJECT_CLASS, configuration, connection);
			uh.test();

			final CohortHandler ch = new CohortHandler(Cohort.OBJECT_CLASS, configuration, connection);
			ch.test();
		} catch (SQLException ex) {
			Db.handleException(ex, "Test error");
		}
	}


	/**
	 * Get latest synchronization token.
	 *
	 * @param objectClass object class
	 * @return synchronization token
	 */
	@Override
	public SyncToken getLatestSyncToken(final ObjectClass objectClass) {
		try {
			if (objectClass.is(Cohort.OBJECT_CLASS_NAME) || objectClass.is(ObjectClass.ALL_NAME)) {
				final CohortHandler h = new CohortHandler(objectClass, configuration, connection);
				return h.getLatestSyncToken();
			} else {
				throw new UnsupportedOperationException("Unsupported object class " + objectClass);
			}
		} catch (SQLException ex) {
			Db.handleException(ex, "Query sync token error");
			return null;
		}
	}


	/**
	 * Get objects changed on the resource.
	 *
	 * @param objectClass object class
	 * @param token synchronization token
	 * @param handler ConnId sync result handler
	 * @param options ConnId operation options
	 */
	@Override
	public void sync(final ObjectClass objectClass, final SyncToken token, final SyncResultsHandler handler, final OperationOptions options) {
		try {
			if (objectClass.is(Cohort.OBJECT_CLASS_NAME) || objectClass.is(ObjectClass.ALL_NAME)) {
				final CohortHandler h = new CohortHandler(objectClass, configuration, connection);
				h.sync(token, handler, options);
			} else {
				throw new UnsupportedOperationException("Unsupported object class " + objectClass);
			}
		} catch (SQLException ex) {
			Db.handleException(ex, "Query sync error");
		}
	}


	/**
	 * Update object on the resource.
	 *
	 * @param objectClass object class
	 * @param uid UID
	 * @param modifications attribute delta set with modifications
	 * @param options ConnId operation options
	 * @return transitively changed attributes
	 */
	@Override
	public Set<AttributeDelta> updateDelta(final ObjectClass objectClass, final Uid uid, final Set<AttributeDelta> modifications, final OperationOptions options) {
		Set<AttributeDelta> changes = null;
		try {
			if (objectClass.is(User.OBJECT_CLASS_NAME)) {
				final UserHandler h = new UserHandler(objectClass, configuration, connection);
				changes = h.updateDelta(uid, modifications, options);
			} else if (objectClass.is(Cohort.OBJECT_CLASS_NAME)) {
				final CohortHandler h = new CohortHandler(objectClass, configuration, connection);
				changes = h.updateDelta(uid, modifications, options);
			} else {
				throw new UnsupportedOperationException("Unsupported object class " + objectClass);
			}
		} catch (SQLException ex) {
			Db.handleException(ex, "Update error");
		}
		return changes;
	}


	/**
	 * Delete object from the resource.
	 *
	 * @param objectClass object class
	 * @param uid UID
	 * @param options ConnId operation options
	 */
	@Override
	public void delete(final ObjectClass objectClass, final Uid uid, final OperationOptions options) {
	try {
			if (objectClass.is(User.OBJECT_CLASS_NAME) || objectClass.is(ObjectClass.ALL_NAME)) {
				final UserHandler h = new UserHandler(objectClass, configuration, connection);
				h.delete(uid, options);
			} else if (objectClass.is(Cohort.OBJECT_CLASS_NAME) || objectClass.is(ObjectClass.ALL_NAME)) {
				final CohortHandler h = new CohortHandler(objectClass, configuration, connection);
				h.delete(uid, options);
			} else {
				throw new UnsupportedOperationException("Unsupported object class " + objectClass);
			}
		} catch (SQLException ex) {
			Db.handleException(ex, "Delete error");
		}
	}

}
