package cz.zcu.connectors.moodle;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.identityconnectors.common.security.GuardedString;
import org.identityconnectors.common.security.SecurityUtil;
import org.identityconnectors.common.logging.Log;
import org.identityconnectors.framework.common.exceptions.InvalidAttributeValueException;
import org.identityconnectors.framework.common.exceptions.UnknownUidException;
import org.identityconnectors.framework.common.objects.Attribute;
import org.identityconnectors.framework.common.objects.AttributeDelta;
import org.identityconnectors.framework.common.objects.AttributeDeltaBuilder;
import org.identityconnectors.framework.common.objects.AttributeInfo;
import org.identityconnectors.framework.common.objects.AttributeInfoBuilder;
import org.identityconnectors.framework.common.objects.ConnectorObject;
import org.identityconnectors.framework.common.objects.Name;
import org.identityconnectors.framework.common.objects.ObjectClass;
import org.identityconnectors.framework.common.objects.ObjectClassInfo;
import org.identityconnectors.framework.common.objects.ObjectClassInfoBuilder;
import org.identityconnectors.framework.common.objects.OperationalAttributes;
import org.identityconnectors.framework.common.objects.OperationOptions;
import org.identityconnectors.framework.common.objects.ResultsHandler;
import org.identityconnectors.framework.common.objects.Schema;
import org.identityconnectors.framework.common.objects.SchemaBuilder;
import org.identityconnectors.framework.common.objects.Uid;

import cz.zcu.connectors.Db;
import cz.zcu.connectors.DbFilter;
import cz.zcu.connectors.moodle.dao.User;


public class UserHandler {
	/** Logging. */
	private static final Log l = Log.getLog(UserHandler.class);

	/** Database table name - users. */
	public static final String TABLE = "mdl_user";
	/** Database table name - attributes . */
	public static final String TABLE_ATTR = "mdl_user_attribute";
	public static final String[] ATTRS_ATTRIBUTES = {
		User.ATTR_PERSONAL_NUMBER,
	};
	public static final String[] ATTRS_ATTRIBUTES_NULLABLE = {
		User.ATTR_PERSONAL_NUMBER,
	};

	/** Connector connection. */
	private final MoodleConnection connection;
	/** Database connection. */
	private final Db db;
	/** Table names according to the current configuration. */
	private final String table, tableAttr;
	private final boolean purge;
	private final String userSuffix;
	private final boolean forcedCleanup;


	/**
	 * User processing class constructor.
	 *
	 * @param objectClass ConnId object class
	 * @param config connector configuration
	 * @param connection connector connection
	 */
	public UserHandler(final ObjectClass objectClass, final MoodleConfiguration config, final MoodleConnection connection) {
		this.connection = connection;
		this.db = connection.db;
		if (config.getTablePrefix().isEmpty()) {
			table = TABLE;
			tableAttr = TABLE_ATTR;
		} else {
			table = config.getTablePrefix() + "." + TABLE;
			tableAttr = config.getTablePrefix() + "." + TABLE_ATTR;
		}
		purge = MoodleConfiguration.DELETE_METHOD_PURGE.equals(config.getDeleteMethod());
		userSuffix = config.getUserSuffix();
		forcedCleanup = config.getForcedCleanup();
	}


	/**
	 * Build schema for User object.
	 *
	 * @return User object schema
	 */
	public final ObjectClassInfo buildSchema() {
		Set<AttributeInfo> attributes;

		attributes = new HashSet<>();
		attributes.add(new AttributeInfoBuilder(Uid.NAME)
				.setNativeName(User.ATTR_ID)
				.setCreateable(false)
				.setRequired(false)
				.setUpdateable(false)
				.build());
		attributes.add(new AttributeInfoBuilder(Name.NAME)
				.setNativeName(User.ATTR_ID)
				.setRequired(true)
				.build());
		attributes.add(new AttributeInfoBuilder(OperationalAttributes.PASSWORD_NAME, GuardedString.class)
				.setNativeName(User.ATTR_PASSWORD)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_AUTH)
				.setRequired(true)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_MNETHOSTID, Long.class)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_IDNUMBER)
			.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_USERNAME)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_FIRSTNAME)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_LASTNAME)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_EMAIL)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_INSTITUTION)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_CONFIRMED, Integer.class)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_DELETED, Integer.class)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_LANG)
				.setRequired(true)
				.build());
		attributes.add(new AttributeInfoBuilder(User.ATTR_PERSONAL_NUMBER)
				.build());

		return new ObjectClassInfoBuilder()
				.setType(User.OBJECT_CLASS_NAME)
				.addAllAttributeInfo(attributes)
				.build();
	}


	/**
	 * Bind parameters from the filter and execute prepared query.
	 *
	 * @param sql SQL command from sqlQuery()
	 * @param st prepared statement from the SQL command
	 * @param filter ConnId filter
	 * @return result set
	 * @throws SQLException SQL Exception
	 */
	private ResultSet dbQuery(final String sql, final PreparedStatement st, final DbFilter filter) throws SQLException {
		if (filter == null) {
			l.ok("SQL: {0}", sql);
		} else if (filter.isValid()) {
			int i = 1;
			List<String> sl = new ArrayList<>();
			for (Object right : filter.right()) {
				if (right == null) {
					throw new InvalidAttributeValueException("Missing value in query filter");
				}
				String v = right.toString();
				st.setString(i++, v);
				sl.add(v);
			}
			l.ok("SQL: {0} [{1}]", sql, String.join(", ", sl));
		} else {
			throw new InvalidAttributeValueException("Empty filter");
		}

		return st.executeQuery();
	}


	/**
	 * Return SQL SELECT command based on the filter.
	 *
	 * @param filter ConnId filter
	 * @return SQL string
	 */
	private String sqlQuery(final DbFilter filter) {
		final String baseSql = "SELECT u.id, auth, mnethostid, idnumber, username, firstname, lastname, email, institution, confirmed, deleted, password, lang, personal_number FROM " + table + " u LEFT JOIN " + tableAttr + " a ON (u.id = a.user_id)";

		if (filter == null) {
			return baseSql;
		} else if (filter.isValid()) {
			return baseSql + " WHERE " + filter.where();
		} else {
			throw new InvalidAttributeValueException("Empty query filter");
		}
	}


	/**
	 * Launch and handle database query.
	 *
	 * @param filter database filter object
	 * @param handler ConnId handler
	 * @param options ConnId operation options
	 */
	public void query(final DbFilter filter, final ResultsHandler handler, final OperationOptions options) throws SQLException {
		l.info("query(query {0}, options {1})", filter, options);
		final DbFilter filter2;

		if (purge || filter != null && (filter.isUid() || filter.isName()))
			// no additional filter: purge delete method, or search by NAME or UID
			filter2 = filter;
		else {
			// additional filter for non-purge delete method
			filter2 = new DbFilter("username NOT LIKE '%" + userSuffix + ".%'", null);
			if (filter != null)
				filter2.and(filter);
		}

		final String sql = sqlQuery(filter2);
		final PreparedStatement st = db.create(sql);

		ResultSet rs = dbQuery(sql, st, filter2);
		while (rs.next()) {
			User user = new User();
			user.load(rs, connection.zone);
			l.ok("fetched user {0} ({1})", user.getUid(), user.getName());
			ConnectorObject co = user.createConnectorObject();
			if (!handler.handle(co)) {
				l.info("listing stopped at {0} ({1})", user.getUid(), user.getName());
				break;
			}
		}
		rs.close();
	}


	private void checkDuplicity(final String uid, final String column, final String value) throws SQLException {
		Utils.checkDuplicity("user", db, table, uid, "id", column, value);
	}


	private void checkUidDuplicity(String uid, String value) throws SQLException {
		checkDuplicity(uid, "id", value);
	}


	private void checkAttrDuplicity(String uid, String column, String value) throws SQLException {
		Utils.checkDuplicity("user attribute", db, tableAttr, uid, "user_id", column, value);
	}


	private int dbUpdate(final String stmtKey, final String sql, final User user, final String oldUid) throws SQLException {
		PreparedStatement st = db.create(stmtKey, sql);
		// use empty string for INSERT
		String defaultValue = oldUid == null ? "" : null;

		int i = 0;
		st.setString(++i, user.getUid());
		st.setString(++i, user.get(User.ATTR_AUTH, defaultValue));
		if (user.get(User.ATTR_MNETHOSTID) == null) {
			st.setNull(++i, java.sql.Types.BIGINT);
		} else {
			st.setLong(++i, user.get(User.ATTR_MNETHOSTID));
		}
		st.setString(++i, user.get(User.ATTR_IDNUMBER, defaultValue));
		st.setString(++i, user.get(User.ATTR_USERNAME, defaultValue));
		st.setString(++i, user.get(User.ATTR_FIRSTNAME, defaultValue));
		st.setString(++i, user.get(User.ATTR_LASTNAME, defaultValue));
		st.setString(++i, user.get(User.ATTR_EMAIL, defaultValue));
		st.setString(++i, user.get(User.ATTR_INSTITUTION, defaultValue));
		if (user.get(User.ATTR_CONFIRMED) == null) {
			st.setNull(++i, java.sql.Types.INTEGER);
		} else {
			st.setInt(++i, user.get(User.ATTR_CONFIRMED));
		}
		if (user.get(User.ATTR_DELETED) == null) {
			st.setNull(++i, java.sql.Types.INTEGER);
		} else {
			st.setInt(++i, user.get(User.ATTR_DELETED));
		}
		if (user.getPassword() == null) {
			st.setNull(++i, java.sql.Types.VARCHAR);
		} else {
			st.setString(++i, SecurityUtil.decrypt(user.getPassword()));
		}
		st.setString(++i, user.get(User.ATTR_LANG));

		if (oldUid == null) {
			l.ok("SQL: {0} [{1}, ...]", sql, String.join(", ", user.getUid(), user.get(User.ATTR_AUTH, defaultValue), String.valueOf((Long)user.get(User.ATTR_MNETHOSTID)), user.get(User.ATTR_USERNAME, defaultValue)));
		} else {
			st.setString(++i, oldUid);
			l.ok("SQL: {0} [{1}, {2}, ..., {3}]", sql, String.join(", ", user.getUid(), user.get(User.ATTR_AUTH, defaultValue), String.valueOf((Long)user.get(User.ATTR_MNETHOSTID)), user.get(User.ATTR_USERNAME, defaultValue)), oldUid);
		}
		st.execute();

		int count = st.getUpdateCount();
		l.ok("SQL: => {0} rows changed", count);

		return count;
	}


	private int dbUpdateMoodleAttributes(final String sql, PreparedStatement st, User user, String oldUid) throws SQLException {
		st.setString(1, user.getUid());
		st.setString(2, user.get(User.ATTR_PERSONAL_NUMBER));
		if (oldUid != null) {
			st.setString(3, oldUid);
			l.ok("SQL: {0} [{1}, {2}, {3}]", sql, user.get(User.ATTR_PERSONAL_NUMBER), user.getUid(), oldUid);
		} else {
			l.ok("SQL: {0} [{1}, {2}]", sql, user.get(User.ATTR_PERSONAL_NUMBER), user.getUid());
		}
		st.executeUpdate();
		int count = st.getUpdateCount();
		l.ok("SQL: => {0} rows changed", count);
		return count;
	}


	/**
	 * Update moodle user attributes
	 *
	 * The record is created, when not exists.
	 * Record is kept even when all attributes are NULL.
	 * No change is performed, when there are no corresponding attributes in the user object, and it is not rename.
	 *
	 * @param oldUid original UID (for rename)
	 * @param user data access object
	 * @param options ConnId update options
	 * @return number of DB records changed
	 * @throws SQLException SQL exception
	 */
	public int updateMoodleAttributes(String oldUid, User user, OperationOptions options) throws SQLException {
		final String insertSql = "INSERT INTO " + tableAttr
				+ "("
					+ "user_id, personal_number"
				+ ") VALUES ("
					+ "?, ?"
				+ ")";
		final String updateSql = "UPDATE " + tableAttr + " SET "
				+ "user_id = " + db.sqlIfNull + "(?, user_id), "
				+ "personal_number = " + db.sqlIfNull + "(?, personal_number) "
				+ "WHERE user_id = ?";
		final String selectSql = "SELECT user_id FROM " + tableAttr + " WHERE user_id = ?";
		ResultSet rs;
		int count = 0;

		if (oldUid == null || oldUid.equals(user.getUid()) && oldUid.equals(user.getName())) {
			boolean hit = false;
			for (String key : ATTRS_ATTRIBUTES) {
				if (user.contains(key)) {
					hit = true;
					break;
				}
			}
			if (!hit)
				return 0;
		}

		PreparedStatement stSelect = db.create(selectSql, selectSql, ResultSet.TYPE_FORWARD_ONLY);
		stSelect.setString(1, oldUid);
		rs = stSelect.executeQuery();
		if (rs.next()) {
			PreparedStatement st = db.create(updateSql);
			dbUpdateMoodleAttributes(updateSql, st, user, oldUid);
			for (String key : ATTRS_ATTRIBUTES_NULLABLE) {
				if (!user.contains(key) || user.get(key) != null)
					continue;
				String sql = "UPDATE " + tableAttr + " SET " + key + " = NULL WHERE user_id = ?";
				PreparedStatement stNull = db.create(sql);
				stNull.setString(1, user.getUid());
				l.ok("SQL: {0} [{1}]", sql, user.getUid());
				stNull.executeUpdate();
				count += st.getUpdateCount();
			}
		} else {
			PreparedStatement st = db.create(insertSql);
			dbUpdateMoodleAttributes(insertSql, st, user, null);
		}
		rs.close();

		return count;
	}


	public Uid create(Set<Attribute> attributes, OperationOptions options) throws SQLException {
		final String sql = "INSERT INTO " + table
			+ " ("
				+ "id, "
				+ "auth, "
				+ "mnethostid, "
				+ "idnumber, "
				+ "username, "
				+ "firstname, "
				+ "lastname, "
				+ "email, "
				+ "institution, "
				+ "confirmed, "
				+ "deleted, "
				+ "password, "
				+ "lang"
			+ ") VALUES ("
				+ "?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?"
			+ ")";
		if (attributes == null || attributes.isEmpty()) {
			l.warn("empty attributes, create request ignored");
			return null;
		}
		User user = new User();
		user.setAttributes(attributes);
		user.validate();

		checkUidDuplicity(null, user.getUid());
		if (user.contains(User.ATTR_USERNAME))
			checkDuplicity(null, User.ATTR_USERNAME, user.get(User.ATTR_USERNAME));
		if (user.contains(User.ATTR_PERSONAL_NUMBER))
			checkAttrDuplicity(null, User.ATTR_PERSONAL_NUMBER, user.get(User.ATTR_PERSONAL_NUMBER));

		dbUpdate("insert user", sql, user, null);

		updateMoodleAttributes(user.getUid(), user, options);

		return new Uid(user.getUid());
	}


	private void cleanupUsername(String oldUid, String userName, OperationOptions options) throws SQLException {
		DbFilter filter = new DbFilter(User.ATTR_USERNAME + " = ?", userName);
		filter.and(User.FIELD_ID + " != ?", oldUid);
		// XXX: dirty hack - this is used for instanse at University of West Bohemia
		filter.and(User.FIELD_ID + " >= 1000000", null);
		final String sql = sqlQuery(filter);
		PreparedStatement st = db.create("cleanup username query", sql);
		st.setString(1, userName);
		st.setString(2, oldUid);
		l.ok("SQL: {0} [{1}, {2}]", sql, userName, oldUid);
		ResultSet rs = st.executeQuery();
		while (rs.next()) {
			String conflictingUid = rs.getString(User.FIELD_ID);
			l.error("conflicting user {0} ({1})", rs.getString(User.ATTR_USERNAME), conflictingUid);
			delete(new Uid(conflictingUid), options);
		}
	}


	public Set<AttributeDelta> updateDelta(Uid uid, Set<AttributeDelta> modifications, OperationOptions options) throws SQLException {
		final String sql = "UPDATE " + table + " SET "
			+ "id = " + db.sqlIfNull + "(?, id), "
			+ "auth = " + db.sqlIfNull + "(?, auth), "
			+ "mnethostid = " + db.sqlIfNull + "(?, mnethostid), "
			+ "idnumber = " + db.sqlIfNull + "(?, idnumber), "
			+ "username = " + db.sqlIfNull + "(?, username), "
			+ "firstname = " + db.sqlIfNull + "(?, firstname), "
			+ "lastname = " + db.sqlIfNull + "(?, lastname), "
			+ "email = " + db.sqlIfNull + "(?, email), "
			+ "institution = " + db.sqlIfNull + "(?, institution), "
			+ "confirmed = " + db.sqlIfNull + "(?, confirmed), "
			+ "deleted = " + db.sqlIfNull + "(?, deleted), "
			+ "password = " + db.sqlIfNull + "(?, password), "
			+ "lang = " + db.sqlIfNull + "(?, lang) "
			+ "WHERE id = ?";
		Set<AttributeDelta> changes = new HashSet<>();
		int count = 0;

		if (uid == null)
			throw new InvalidAttributeValueException("UID is null");
		String oldUid = uid.getUidValue();

		if (modifications == null || modifications.isEmpty()) {
			l.warn("empty attributes, updateDelta request ignored");
			return null;
		}
		User user = new User();
		user.setDelta(modifications);

		// check and prepare for rename (can be through both - NAME attribute and UID, UID used in dbUpdate())
		if (user.getName() != null && !user.getName().equals(oldUid))
			user.setUid(user.getName());
		if (user.getUid() != null && !user.getUid().equals(oldUid)) {
			checkUidDuplicity(oldUid, user.getUid());
			changes.add(AttributeDeltaBuilder.build(Uid.NAME, user.getUid()));
		}

		if (user.contains(User.ATTR_USERNAME)) {
			if (forcedCleanup)
				cleanupUsername(oldUid, user.get(User.ATTR_USERNAME), options);
			checkDuplicity(oldUid, User.ATTR_USERNAME, user.get(User.ATTR_USERNAME));
		}
		if (user.contains(User.ATTR_PERSONAL_NUMBER))
			checkAttrDuplicity(oldUid, User.ATTR_PERSONAL_NUMBER, user.get(User.ATTR_PERSONAL_NUMBER));

		// update
		count += dbUpdate("update user", sql, user, oldUid);
		if (count == 0)
			throw new UnknownUidException("User " + oldUid + " to update doesn't exist");

		if (user.getUid() == null)
			user.setUid(oldUid);

		// explicit NULL needed
		List<String> nullValues = new ArrayList<>(1);
		List<GuardedString> nullPwdValues = new ArrayList<>(1);
		nullValues.add(null);
		nullPwdValues.add(null);
		for (String key : User.ATTRS.keySet()) {
			if (user.contains(key) && !User.ATTR_PERSONAL_NUMBER.equals(key)) {
				// background database schema forbids NULL, only empty strings or 0 are allowed
				// ==> reset to empty string in DB
				// ==> reset when null or empty string
				// ==> forbid reset of non-string
				Object v = user.get(key);
				boolean isString = User.ATTRS.containsKey(key) && (User.ATTRS.get(key) == String.class);
				boolean isPassword = User.ATTRS.containsKey(key) && (User.ATTRS.get(key) == GuardedString.class);
				boolean isEmptyString = isString && (v != null && ((String)v).isEmpty());
				if (v == null && !isString && !isPassword)
					throw new InvalidAttributeValueException("Cannot reset non-string attribute '" + key + "'");
				if (v != null && !isEmptyString)
					continue;
				final String sqlNull = "UPDATE " + table + " SET " + key + " = '' WHERE id = ?";
				PreparedStatement st = db.create(sqlNull);
				st.setString(1, user.getUid());
				l.ok("SQL: {0} [{1}]", sqlNull, user.getUid());
				st.execute();
				count += st.getUpdateCount();
				// if empty string is used, send back to IdM the according update to null
				if (isEmptyString)
					changes.add(AttributeDeltaBuilder.build(key, isPassword ? nullPwdValues : nullValues));
			}
		}

		l.info("{0} updates", count);

		updateMoodleAttributes(oldUid, user, options);

		return changes.isEmpty() ? null : changes;
	}


	public int delete(Uid uid, OperationOptions options) throws SQLException {
		final String sqlUserPurge = "DELETE FROM " + table + " WHERE id = ?";
		final String sqlUserRecycle = "UPDATE " + table + " SET username=CONCAT(username, '.', unix_timestamp()), email=IF(email IS NULL OR email='', email, CONCAT(email, '.', unix_timestamp())) WHERE id = ? AND username NOT LIKE '%" + userSuffix + ".%'";
		final String sqlAttrPurge = "DELETE FROM " + tableAttr + " WHERE user_id = ?";
		final String sqlAttrRecycle = "UPDATE " + tableAttr + " SET personal_number = NULL WHERE user_id = ?";
		final String sql = purge ? sqlUserPurge : sqlUserRecycle;
		final String sqlAttr = purge ? sqlAttrPurge : sqlAttrRecycle;
		PreparedStatement st;

		if (uid == null)
			throw new InvalidAttributeValueException("UID is null");
		String uidValue = uid.getUidValue();

		st = db.create(sqlAttr);
		st.setString(1, uidValue);
		l.ok("SQL: {0} [{1}]", sqlAttr, uidValue);
		st.execute();
		int countAttr = st.getUpdateCount();
		l.ok("SQL delete count {0}", countAttr);

		st = db.create(sql);
		st.setString(1, uidValue);
		l.ok("SQL: {0} [{1}]", sql, uidValue);
		st.execute();
		int countUser = st.getUpdateCount();
		l.ok("SQL delete count {0}", countUser);

		if (countUser == 0 && countAttr == 0)
			throw new UnknownUidException("User " + uid.getUidValue() + " to delete doesn't exist");

		return countUser + countAttr;
	}


	public void test() throws SQLException {
		final String[] sqls = {
			"SELECT id FROM " + table,
			"SELECT id, user_id FROM " + tableAttr,
		};
		for (final String sql: sqls) {
			PreparedStatement st = db.create(sql, sql, ResultSet.TYPE_FORWARD_ONLY);
			l.ok("SQL: {0}", sql);
			ResultSet rs = st.executeQuery();
			rs.next();
			rs.close();
		}
	}
}
