package cz.zcu.connectors.moodle.dao;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.Instant;
import java.time.ZonedDateTime;
import java.time.ZoneId;
import java.util.HashMap;
import java.util.Map;

import org.identityconnectors.common.security.GuardedString;
import org.identityconnectors.framework.common.exceptions.InvalidAttributeValueException;
import org.identityconnectors.framework.common.objects.ConnectorObjectBuilder;
import org.identityconnectors.framework.common.objects.ObjectClass;
import org.identityconnectors.framework.common.objects.Name;
import cz.zcu.connectors.dao.Generic;


/**
 * Cohort data access object.
 */
public class Cohort extends Generic {
	/** Object class name. */
	public static final String OBJECT_CLASS_NAME = "Cohort";
	/** Object class. */
	public static final ObjectClass OBJECT_CLASS = new ObjectClass(OBJECT_CLASS_NAME);

	/** "id" attribute (UID operational attribute). */
	public static final String ATTR_ID = "cohort_id";
	/** "name" attribute (NAME operational attribute). */
	public static final String ATTR_NAME = "name";
	/** "contextid" attribute. */
	public static final String ATTR_CONTEXTID = "contextid";
	/** "idnumber" attribute. */
	public static final String ATTR_IDNUMBER = "idnumber";
	/** "description" attribute. */
	public static final String ATTR_DESCRIPTION = "description";
	/** "descriptionformat" attribute. */
	public static final String ATTR_DESCRIPTIONFORMAT = "descriptionformat";
	/** "visible" attribute. */
	public static final String ATTR_VISIBLE = "visible";
	/** "component" attribute. */
	public static final String ATTR_COMPONENT = "component";
	/** "theme" attribute. */
	public static final String ATTR_THEME = "theme";
	/** "timecreated" attribute. */
	public static final String ATTR_TIMECREATED = "timecreated";
	/** "timemodified" attribute. */
	public static final String ATTR_TIMEMODIFIED = "timemodified";

	/** ID database field. */
	public static final String FIELD_ID = "id";
	/** NAME database field. */
	public static final String FIELD_NAME = "name";

	/** Attributes specification. */
	public static final Map<String, Class<?>> ATTRS = new HashMap<>();


	/**
	 * Resource data access object constructor.
	 */
	public Cohort() {
		super(ATTR_NAME, ATTR_ID, ATTRS);
	}


	/**
	 * Resource data access object as clone constructor.
	 *
	 * @param o original object to clone
	 */
	public Cohort(final Cohort o) {
		super(o);
	}


	/**
	 * Clone resource data access object.
	 *
	 * @return cloned object
	 */
	public Cohort clone() {
		return new Cohort(this);
	}


	/**
	 * Construct ConnId connector object builder from current data.
	 */
	@Override
	public ConnectorObjectBuilder connectorObjectBuilder() {
		return connectorObjectBuilder(new ObjectClass(OBJECT_CLASS_NAME));
	}


	/**
	 * Load data from database result set.
	 *
	 * @param rs database result set
	 * @param zone zone id
	 */
	public void load(final ResultSet rs, final ZoneId zone) throws SQLException {
		for (String attrName : ATTRS.keySet()) {
			Class<?> type = ATTRS.get(attrName);
			Object value = null;
			String key = attrName;

			// avoid using "id" externally due to midPoint limitation
			if (ATTR_ID.equals(attrName)) {
				key = FIELD_ID;
			}

			if (type == String.class) {
				value = rs.getString(key);
			} else if (type == Integer.class) {
				value = rs.getInt(key);
			} else if (type == Long.class) {
				value = rs.getLong(key);
			} else if (type == ZonedDateTime.class) {
				// in database always stored as bigint with precision to seconds
				Long t = rs.getLong(key);
				value = ZonedDateTime.ofInstant(Instant.ofEpochSecond(t), zone);
			} else if (type == GuardedString.class) {
				String s = rs.getString(key);
				if (s != null) {
					value = new GuardedString(s.toCharArray());
				}
			} else if (type == Boolean.class) {
				String s = rs.getString(key);
				if (s != null) {
					value = "A".equals(s) || "1".equals(s);
				}
			}
			if (value != null && !rs.wasNull()) {
				this.getData().put(attrName, value);
			}
		}

		if (getEnableAttr() != null) {
			setEnable(this.get(getEnableAttr()));
			// duplicity with __ENABLE__ attribute
			this.remove(getEnableAttr());
		}

		refresh();
	}


	/**
	 * Validate data in resource object.
	 *
	 * @see cz.zcu.connectors.dao.Generic#validate()
	 */
	@Override
	public void validate() {
		if (getName() == null) {
			throw new InvalidAttributeValueException("missing " + Name.NAME + " attribute in cohort (" + ATTR_NAME + ")");
		}
	}


	static {
		ATTRS.put(ATTR_ID, String.class);
		ATTRS.put(ATTR_NAME, String.class);
		ATTRS.put(ATTR_CONTEXTID, Long.class);
		ATTRS.put(ATTR_IDNUMBER, String.class);
		ATTRS.put(ATTR_DESCRIPTION, String.class);
		ATTRS.put(ATTR_DESCRIPTIONFORMAT, Integer.class);
		ATTRS.put(ATTR_VISIBLE, Integer.class);
		ATTRS.put(ATTR_COMPONENT, String.class);
		ATTRS.put(ATTR_THEME, String.class);
		ATTRS.put(ATTR_TIMECREATED, ZonedDateTime.class);
		ATTRS.put(ATTR_TIMEMODIFIED, ZonedDateTime.class);
	}
}
