package cz.zcu.connectors.moodle.dao;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.ZonedDateTime;
import java.time.ZoneId;
import java.util.HashMap;
import java.util.Map;

import org.identityconnectors.common.security.GuardedString;
import org.identityconnectors.framework.common.exceptions.InvalidAttributeValueException;
import org.identityconnectors.framework.common.objects.ConnectorObjectBuilder;
import org.identityconnectors.framework.common.objects.ObjectClass;
import org.identityconnectors.framework.common.objects.Uid;
import cz.zcu.connectors.dao.Generic;


/**
 * User data access object.
 */
public class User extends Generic {
	/** Object class name. */
	public static final String OBJECT_CLASS_NAME = "User";
	/** Object class. */
	public static final ObjectClass OBJECT_CLASS = new ObjectClass(OBJECT_CLASS_NAME);

	/** "id" attribute (UID operational attribute). */
	public static final String ATTR_ID = "user_id";
	public static final String ATTR_AUTH = "auth";
	public static final String ATTR_MNETHOSTID = "mnethostid";
	public static final String ATTR_IDNUMBER = "idnumber";
	public static final String ATTR_USERNAME = "username";
	public static final String ATTR_FIRSTNAME = "firstname";
	public static final String ATTR_LASTNAME = "lastname";
	public static final String ATTR_EMAIL = "email";
	public static final String ATTR_INSTITUTION = "institution";
	public static final String ATTR_CONFIRMED = "confirmed";
	public static final String ATTR_DELETED = "deleted";
	public static final String ATTR_PASSWORD = "password"; // gitleaks:allow
	public static final String ATTR_LANG = "lang";
	public static final String ATTR_PERSONAL_NUMBER = "personal_number";

	/** ID database field. */
	public static final String FIELD_ID = "u.id";

	/** Attributes specification. */
	public static final Map<String, Class<?>> ATTRS = new HashMap<>();


	/**
	 * Resource data access object constructor.
	 */
	public User() {
		super(ATTR_ID, ATTR_ID, ATTRS);
		setPasswordAttr(ATTR_PASSWORD);
	}


	/**
	 * Resource data access object as clone constructor.
	 *
	 * @param o original object to clone
	 */
	public User(User o) {
		super(o);
	}


	/**
	 * Clone resource data access object.
	 *
	 * @return cloned object
	 */
	public User clone() {
		return new User(this);
	}


	/**
	 * Construct ConnId connector object builder from current data.
	 */
	@Override
	public ConnectorObjectBuilder connectorObjectBuilder() {
		return connectorObjectBuilder(new ObjectClass(OBJECT_CLASS_NAME));
	}


	/**
	 * Load data from database result set.
	 *
	 * @param rs database result set
	 * @param zone zone id
	 */
	public void load(ResultSet rs, final ZoneId zone) throws SQLException {
		for (String attrName : ATTRS.keySet()) {
			Class<?> type = ATTRS.get(attrName);
			Object value = null;
			String key = attrName;

			// avoid using "id" externally due to midPoint limitation
			if (ATTR_ID.equals(attrName)) {
				key = FIELD_ID;
			}

			if (type == String.class) {
				String s = rs.getString(key);
				if (s != null && !s.isEmpty()) {
					value = s;
				}
			} else if (type == Integer.class) {
				value = rs.getInt(key);
			} else if (type == Long.class) {
				value = rs.getLong(key);
			} else if (type == ZonedDateTime.class) {
				value = sqlToZDT(rs.getTimestamp(key), zone);
			} else if (type == GuardedString.class) {
				String s = rs.getString(key);
				if (s != null && !s.isEmpty()) {
					value = new GuardedString(s.toCharArray());
				}
			} else if (type == Boolean.class) {
				String s = rs.getString(key);
				if (s != null) {
					value = "A".equals(s);
				}
			}
			if (value != null && !rs.wasNull()) {
				this.getData().put(attrName, value);
			}
		}

		refresh();
	}


	/**
	 * Validate data in resource object.
	 *
	 * @see cz.zcu.connectors.dao.Generic#validate()
	 */
	@Override
	public void validate() {
		if (getUid() == null) {
			throw new InvalidAttributeValueException("missing " + Uid.NAME + " attribute in user (" + ATTR_ID + ")");
		}
	}


	static {
		ATTRS.put(ATTR_ID, Long.class);
		ATTRS.put(ATTR_AUTH, String.class);
		ATTRS.put(ATTR_MNETHOSTID, Long.class);
		ATTRS.put(ATTR_IDNUMBER, String.class);
		ATTRS.put(ATTR_USERNAME, String.class);
		ATTRS.put(ATTR_FIRSTNAME, String.class);
		ATTRS.put(ATTR_LASTNAME, String.class);
		ATTRS.put(ATTR_EMAIL, String.class);
		ATTRS.put(ATTR_INSTITUTION, String.class);
		ATTRS.put(ATTR_CONFIRMED, Integer.class);
		ATTRS.put(ATTR_DELETED, Integer.class);
		ATTRS.put(ATTR_PASSWORD, GuardedString.class);
		ATTRS.put(ATTR_LANG, String.class);
		ATTRS.put(ATTR_PERSONAL_NUMBER, String.class);
	}
}
