package cz.zcu.connectors.moodle;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import cz.zcu.connectors.moodle.dao.Cohort;
import org.identityconnectors.common.logging.Log;
import org.identityconnectors.common.security.GuardedString;
import org.identityconnectors.common.security.SecurityUtil;
import org.identityconnectors.framework.api.APIConfiguration;
import org.identityconnectors.framework.api.ConnectorFacade;
import org.identityconnectors.framework.api.ConnectorFacadeFactory;
import org.identityconnectors.framework.common.objects.Attribute;
import org.identityconnectors.framework.common.objects.AttributeDelta;
import org.identityconnectors.framework.common.objects.AttributeDeltaBuilder;
import org.identityconnectors.framework.common.objects.ConnectorObject;
import org.identityconnectors.framework.common.objects.Name;
import org.identityconnectors.framework.common.objects.ObjectClass;
import org.identityconnectors.framework.common.objects.OperationOptionsBuilder;
import org.identityconnectors.framework.common.objects.SyncDelta;
import org.identityconnectors.framework.common.objects.SyncResultsHandler;
import org.identityconnectors.framework.common.objects.SyncToken;
import org.identityconnectors.framework.common.objects.Uid;
import org.identityconnectors.framework.common.objects.filter.Filter;
import org.identityconnectors.framework.common.objects.filter.FilterBuilder;
import org.identityconnectors.framework.spi.Connector;
import org.identityconnectors.test.common.PropertyBag;
import org.identityconnectors.test.common.TestHelpers;
import org.identityconnectors.test.common.ToListResultsHandler;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.Assert;
import org.testng.SkipException;

import cz.zcu.connectors.dao.Generic;


public class MoodleConnectorTestBase {
	/** Logging. */
	private static final Log l = Log.getLog(MoodleConnectorTestBase.class);
	/** Testsuite properties. */
	private static final PropertyBag PROPERTIES = TestHelpers.getProperties(MoodleConnector.class, null);
	/** Connector facade for testing. */
	private volatile ConnectorFacade connectorFacade;
	/** Object class. */
	private final ObjectClass objectClass;


	/**
	 * Test base class constructor.
	 *
	 * @param objectClass object class
	 */
	public MoodleConnectorTestBase(final ObjectClass objectClass) {
		this.objectClass = objectClass;
	}


	/**
	 * Tests setup.
	 */
	@BeforeClass
	public void setUp() {
		MoodleConfiguration config = new MoodleConfiguration();
		Map<String, Object> configData = new HashMap<>();
		for (String k : Arrays.asList("jdbcDriver", "jdbcUrl", "jdbcUser", "deleteMethod")) {
			try {
				Object v;
				switch(k) {
				default:
					v = PROPERTIES.getProperty("configuration." + k, String.class);
				}
				if (v != null) {
					configData.put(k, v);
				}
			} catch (IllegalArgumentException iae) {
				// OK
			}
		}
		TestHelpers.fillConfiguration(config, configData);
		for (String k : Arrays.asList("jdbcPassword")) {
			try {
				GuardedString pwd;
				switch(k) {
				case "jdbcPassword":
					pwd = PROPERTIES.getProperty("configuration." + k, GuardedString.class);
					config.setJdbcPassword(pwd);
					break;
				default:
				}
			} catch (IllegalArgumentException iae) {
				// OK
			}
		}
		// db = new Db("moodle", config);
	}


	/**
	 * Tests teardown.
	 */
	@AfterClass
	public void tearDown() {
		if (connectorFacade != null) {
			connectorFacade.dispose();
		}
	}


	/**
	 * Setup for each test.
	 */
	@BeforeMethod
	public void setUpEach() {
	}


	/**
	 * Create object on the resource for testing using the connector.
	 *
	 * @param facade connector
	 * @param o data object
	 * @return UID of the created object
	 */
	protected Uid create(final ConnectorFacade facade, final Generic o) {
		delete(facade, o);
		return facade.create(objectClass, o.getAttributes(), null);
	}


	/**
	 * Delete object on the resource using the connector.
	 *
	 * @param facade connector
	 * @param o data object
	 */
	protected void delete(final ConnectorFacade facade, final Generic o) {
		final ToListResultsHandler handler = new ToListResultsHandler();

		facade.search(objectClass, FilterBuilder.equalTo(new Name(o.getName())), handler, null);
		if (handler.getObjects() != null && !handler.getObjects().isEmpty()) {
			facade.delete(objectClass, handler.getObjects().get(0).getUid(), null);
		}
	}


	/**
	 * Load data object from testsuite properties.
	 *
	 * @param o data object to fill
	 * @param name test object name
	 * @param ATTRS attributes specification
	 */
	protected void load(final Generic o, final String name, final Map<String, Class<?>> ATTRS) {
		final String prefix = name == null || "".equals(name) ? "" : name + ".";

		for (final String key : ATTRS.keySet()) {
			final Class<?> type = ATTRS.get(key);
			final Object value = PROPERTIES.getProperty(prefix + key, type, null);
			if (value != null) {
				o.getData().put(key, value);
			}
		}
	}



	/**
	* Get object test.
	*
	* @param o data object
	*/
	protected void getTest(final Generic o) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();

		Uid uid = new Uid(o.getUid());

		builder.setAttributesToGet(Name.NAME);
		ConnectorObject co = facade.getObject(objectClass, uid, builder.build());
		check(co, o);
	}


	/**
	 * Get unknown object test.
	 *
	 * @param o data object
	 */
	protected void getUnknownTest(final Generic o) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();
		ConnectorObject co;

		builder.setAttributesToGet(Name.NAME);
		co = facade.getObject(objectClass, new Uid(o.getUid()), builder.build());
		Assert.assertNull(co);
	}



	/**
	 * Search exact test.
	 *
	 * @param o data object, which must exist on the resource
	 * @param filters filters to use
	 */
	protected void searchExactTest(final Generic o, final Filter[] filters) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();
		ToListResultsHandler handler;
		ConnectorObject co;

		for (Filter f : filters) {
			handler = new ToListResultsHandler();
			facade.search(objectClass, f, handler, builder.build());
			Assert.assertEquals(handler.getObjects().size(), 1, "exactly one object found");
			co = handler.getObjects().get(0);
			check(co, o);
		}
	}


	/**
	 * Search all test.
	 *
	 * @param o data object, which must exist on the resource
	 */
	protected void searchAllTest(final Generic o) {
		final int pageSize = 10;
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();
		builder.setPageSize(pageSize);

		final ToListResultsHandler handler = new ToListResultsHandler();
		facade.search(objectClass, null, handler, builder.build());
		Assert.assertTrue(handler.getObjects().size() >= 2, "objects found");
		Assert.assertTrue(handler.getObjects().stream().anyMatch(it -> o.getUid().equals(it.getUid().getUidValue())), "test object found");
	}


	/**
	 * Create test.
	 *
	 * @param o data object for create
	 */
	protected void createTest(final Generic o) {
		final ConnectorFacade facade = getFacade(null);
		ConnectorObject co;

		Uid uid = create(facade, o);
		Assert.assertNotNull(uid, "returned __UID__ from create");
		Assert.assertNotNull(uid.getUidValue(), "returned __UID__ value from create");
		// compare returned UID only when explicitly specified
		if (o.getUid() != null) {
			Assert.assertEquals(uid.getUidValue(), o.getUid(), "returned __UID__ value from create");
		}

		co = facade.getObject(objectClass, uid, null);
		check(co, o);
	}


	/**
	 * Delete test.
	 *
	 * @param o data object to delete (will be created first)
	 */
	protected void deleteTest(final Generic o) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();
		ConnectorObject co;

		Uid uid = create(facade, o);

		facade.delete(objectClass, uid, builder.build());
		co = facade.getObject(objectClass, uid, builder.build());
		Assert.assertNull(co, "object really deleted");
	}


	/**
	 * Delete unknown test.
	 *
	 * @param o data object, which doesn't exist on the resource
	 */
	protected void deleteUnknownTest(final Generic o) {
		final ConnectorFacade facade = getFacade(null);

		facade.delete(objectClass, new Uid(o.getUid()), null);
	}


	/**
	 * Check related changes after changing UID attribute.
	* @param changes changes to check
	* @param newUid new UID
	 */
	private void checkChanges(final Set<AttributeDelta> changes, final String newUid) {
		Assert.assertNotNull(changes, "changes");
		Assert.assertEquals(changes.size(), 1, "changes has 1 element");
		AttributeDelta ad = changes.iterator().next();
		Assert.assertEquals(ad.getName(), Uid.NAME, "changes UID attribute");
		Assert.assertNotNull(ad.getValuesToReplace(), "changes UID values");
		Assert.assertEquals(ad.getValuesToReplace().get(0), newUid, "changes UID value");
	}


	/**
	 * Update test.
	 *
	 * @param o original data
	 * @param o2 new data with updated values
	 * @param withCreate create the original object on the resource first
	 * @param withRename perform and test renaming too
	 */
	protected void updateTest(final Generic o, final Generic o2, final boolean withCreate, final boolean withRename) {
		final ConnectorFacade facade = getFacade(null);
		ConnectorObject co;
		Set<AttributeDelta> modifications;

		delete(facade, o);
		delete(facade, o2);
		String uidOrig = o.getUid();
		if (withCreate) {
			uidOrig = create(facade, o).getUidValue();
		}
		Generic oNew = o2.clone();
		oNew.setUid(uidOrig);
		oNew.set(oNew.getUidAttr(), uidOrig);
		if (!withRename) {
			oNew.setName(o.getName());
			oNew.set(oNew.getNameAttr(), o.get(oNew.getNameAttr()));
		}

		modifications = oNew.getDelta();
		facade.updateDelta(objectClass, new Uid(uidOrig), modifications, null);
		co = facade.getObject(objectClass, new Uid(oNew.getUid()), null);
		check(co, oNew);
	}


	/**
	 * Rename UID test.
	 *
	 * @param o original data
	 * @param o2 data with new values
	 */
	protected void renameUidTest(final Generic o, final Generic o2) {
		if (o2.getUid() == null) {
			throw new SkipException("no rename");
		}

		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();

		delete(facade, o);
		delete(facade, o2);
		String oldUid = create(facade, o).getUidValue();

		String uidAttr = o.getUidAttr();
		for (String newUid : Arrays.asList(oldUid, o2.getUid())) {
			l.info("Renaming {0} -> {1}", oldUid, newUid);
			Set<AttributeDelta> modifications = new HashSet<>();

			modifications.add(AttributeDeltaBuilder.build(uidAttr, newUid));
			Set<AttributeDelta> changes = facade.updateDelta(objectClass, new Uid(oldUid), modifications, builder.build());
			if (oldUid.equals(newUid)) {
				Assert.assertNull(changes, "changes must be null on identical rename");
			} else {
				checkChanges(changes, newUid);
			}

			ConnectorObject co = facade.getObject(objectClass, new Uid(newUid), null);
			Assert.assertNotNull(co, objectClass + " found by the new UID");
			Assert.assertEquals(co.getUid().getUidValue(), newUid, "UID the new UID");

			Generic oNew = o.clone();
			oNew.set(uidAttr, newUid);
			oNew.refresh();
			check(co, oNew);
		}
		ConnectorObject co = facade.getObject(objectClass, new Uid(oldUid), null);
		Assert.assertNull(co, objectClass + " properly not found by the old UID");
	}


	/**
	 * Rename test.
	 *
	 * @param o original data object
	 * @param o2 object with new name
	 */
	protected void renameTest(final Generic o, final Generic o2) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();

		delete(facade, o);
		delete(facade, o2);
		Uid uid = create(facade, o);

		String name = o.getName();
		for (String newName : Arrays.asList(name, o2.getName())) {
			l.info("Renaming {0} -> {1}", name, newName);
			Set<AttributeDelta> modifications = new HashSet<>();

			modifications.add(AttributeDeltaBuilder.build(Name.NAME, newName));
			Set<AttributeDelta> changes = facade.updateDelta(objectClass, uid, modifications, builder.build());
			Assert.assertNull(changes, "changes must be always null on rename without changing uid");

			ToListResultsHandler handler = new ToListResultsHandler();
			facade.search(objectClass, FilterBuilder.equalTo(new Name(newName)), handler, null);
			Assert.assertNotNull(handler.getObjects(), "any objects found by new name");
			Assert.assertEquals(handler.getObjects().size(), 1, "exactly one object found by new name");
			ConnectorObject co = handler.getObjects().get(0);
			Assert.assertNotNull(co, "object found by the new name");
			Assert.assertEquals(co.getName().getNameValue(), newName, "NAME is the new name");

			Generic oNew = o.clone();
			oNew.setName(newName);
			check(co, oNew);
		}
	}


	/**
	 * Rename exits tests.
	 *
	 * @param o original object
	 * @param o2 data object with conflicting name already exiting on the resource
	 */
	protected void renameExistsTest(final Generic o, final Generic o2) {
		final ConnectorFacade facade = getFacade(null);

		delete(facade, o);
		delete(facade, o2);
		Uid uid = create(facade, o);
		create(facade, o2);

		Set<AttributeDelta> modifications = new HashSet<>();
		modifications.add(AttributeDeltaBuilder.build(Name.NAME, o2.getName()));
		facade.updateDelta(objectClass, uid, modifications, null);
	}


	/**
	 * Update reset test.
	 *
	 * @param o original object
	 * @param ATTRS attributes to reset
	 */
	protected void updateResetTest(final Generic o, final String... ATTRS) {
		final ConnectorFacade facade = getFacade(null);
		ConnectorObject co;
		Set<AttributeDelta> modifications = new HashSet<>();

		delete(facade, o);
		Uid uid = create(facade, o);

		List<Object> nullValue = new ArrayList<>();
		nullValue.add(null);
		for (String attr : ATTRS) {
			modifications.add(AttributeDeltaBuilder.build(attr, nullValue));
		}
		Set<AttributeDelta> changes = facade.updateDelta(objectClass, uid, modifications, null);
		Assert.assertNull(changes, "changes");

		co = facade.getObject(objectClass, uid, null);
		for (String attr : ATTRS) {
			Attribute a = co.getAttributeByName(attr);
			Assert.assertNull(a, attr + " attribute");
		}
	}


	/**
	 * Short sleep (used for testing lifeSync).
	 */
	private void sleepForSync() {
		final int sleepTime = 1500;
		try {
			// time precision of the time stamp in the database is millisecond,
			// ==> wait one millisecond
			// some databases (MSSQL) has limited time quantum
			// => wait some more
			l.ok("wait start: {0}", (new java.util.Date()).getTime());
			Thread.sleep(sleepTime);
			l.ok("wait end:   {0}", (new java.util.Date()).getTime());
		} catch (InterruptedException e) { }
	}


	/**
	 * Touch object on the resource.
	 *
	 * @param facade connector
	 * @param objectClassTouch object class
	 * @param uid object UID to touch
	 * @param o original data object
	 * @param o2 data object with new value to use
	 * @param oEmpty helper object to use for updates
	 * @param attrName attribute name to use for touching
	 */
	private void touch(final ConnectorFacade facade, final ObjectClass objectClassTouch, final Uid uid, final Generic o, final Generic o2,
			final Generic oEmpty, final String attrName) {
		oEmpty.setUid(uid.getUidValue());
		oEmpty.set(attrName, o2.get(attrName));
		facade.updateDelta(objectClassTouch, uid, oEmpty.getDelta(), null);
		oEmpty.set(attrName, o.get(attrName));
		facade.updateDelta(objectClassTouch, uid, oEmpty.getDelta(), null);
	}


	/**
	 * Get latest sync test.
	 *
	 * @param objectClassFetch object class to use for using lifeSync
	 * @param objectClassTouch object class to use for changing the resource object
	 * @param o original data object to change
	 * @param o2 data object with new value to use
	 * @param oEmpty helper object to use for updates
	 * @param attrName attribute name to usr for touching
	 */
	protected void getLatestSyncTest(final ObjectClass objectClassFetch, final ObjectClass objectClassTouch, final Generic o, final Generic o2,
			final Generic oEmpty, final String attrName) {
		final ConnectorFacade facade = getFacade(null);

		delete(facade, o);
		delete(facade, o2);

		SyncToken tokenOld = facade.getLatestSyncToken(objectClassFetch);
		Assert.assertNotNull(tokenOld, "initial token");

		sleepForSync();

		Uid uid = create(facade, o);
		touch(facade, objectClassTouch, uid, o, o2, oEmpty, attrName);

		sleepForSync();

		SyncToken tokenNew = facade.getLatestSyncToken(objectClassFetch);
		Assert.assertNotNull(tokenNew, "new token");
		l.info("old token: {0}", tokenOld.getValue());
		l.info("new token: {0}", tokenNew.getValue());
		Assert.assertTrue((Long) tokenNew.getValue() > (Long) tokenOld.getValue(), "new token > old token");
	}


	/**
	 * Synchronization test.
	 *
	 * @param objectClassFetch object class to use for using lifeSync
	 * @param objectClassTouch object class to use for changing the resource object
	 * @param o original data object to change
	 * @param o2 data object with new value to use
	 * @param oEmpty helper object to use for updates
	 * @param attrName attribute name to usr for touching
	 */
	protected void syncTest(final ObjectClass objectClassFetch, final ObjectClass objectClassTouch, final Generic o, final Generic o2,
			final Generic oEmpty, final String attrName) {
		final ConnectorFacade facade = getFacade(null);
		List<SyncDelta> deltas = new LinkedList<>();
		ConnectorObject co;

		delete(facade, o);
		delete(facade, o2);

		SyncToken initialToken = facade.getLatestSyncToken(objectClassFetch);
		sleepForSync();

		// the first account
		l.info("first update");
		Uid uid = create(facade, o);
		touch(facade, objectClassTouch, uid, o, o2, oEmpty, attrName);
		sleepForSync();

		facade.sync(objectClassFetch, initialToken, new SyncResultsHandler() {
			@Override
			public boolean handle(final SyncDelta delta) {
				deltas.add(delta);
				return true;
			}
		}, null);
		Assert.assertTrue(deltas.size() >= 1, "any objects found from first update");
		Optional<SyncDelta> od = deltas.stream()
				.filter(delta -> uid.getUidValue().equals(delta.getObject().getUid().getUidValue()))
				.findFirst();
		Assert.assertTrue(od.isPresent(), "our object found from first update");
		co = od.get().getObject();
		check(co, o);
		SyncToken newToken = od.get().getToken();
		l.info("initial token: {0}", initialToken.getValue());
		l.info("first token:   {0}", newToken.getValue());
		Assert.assertTrue(((Long) newToken.getValue()) > ((Long) initialToken.getValue()), "first token > initial token");

		// the second update
		l.info("second update");
		touch(facade, objectClassTouch, uid, o, o2, oEmpty, attrName);
		sleepForSync();

		SyncToken oldToken = newToken;
		deltas.clear();
		facade.sync(objectClassFetch, oldToken, new SyncResultsHandler() {
			@Override
			public boolean handle(final SyncDelta delta) {
				deltas.add(delta);
				return true;
			}
		}, null);
		Assert.assertTrue(deltas.size() >= 1, "any objects found from second update");
		od = deltas.stream()
				.filter(delta -> uid.getUidValue().equals(delta.getObject().getUid().getUidValue()))
				.findFirst();
		Assert.assertTrue(od.isPresent(), "our object found from second update");
		co = od.get().getObject();
		check(co, o);
		newToken = od.get().getToken();
		l.info("initial token: {0}", initialToken.getValue());
		l.info("first token:   {0}", oldToken.getValue());
		l.info("second token:  {0}", newToken.getValue());
		Assert.assertTrue(((Long) newToken.getValue()) > ((Long) initialToken.getValue()), "second token > first token");

		// no remaining objects
		l.info("remaining check");
		oldToken = newToken;
		deltas.clear();
		facade.sync(objectClassFetch, oldToken, new SyncResultsHandler() {
			@Override
			public boolean handle(final SyncDelta delta) {
				deltas.add(delta);
				return true;
			}
		}, null);
		Assert.assertEquals(deltas.size(), 0, "no remaining objects");
	}


	/**
	 * Decrypt and compare passwords.
	 *
	 * @param attrName name of the attribute (for debugging only)
	 * @param values tested value
	 * @param expectedPassword encrypted expected value
	 */
	private void checkPwd(final String attrName, final List<Object> values, final GuardedString expectedPassword) {
		final String msg = attrName + " guarded attribute value";

		if (values == null || values.isEmpty()) {
			Assert.assertNull(expectedPassword, msg);
			return;
		}
		final GuardedString foundPassword = (GuardedString) values.get(0);
		if (foundPassword == null || expectedPassword == null) {
			Assert.assertEquals(foundPassword, expectedPassword, msg);
		} else {
			Assert.assertEquals(SecurityUtil.decrypt(foundPassword), SecurityUtil.decrypt(expectedPassword), msg);
		}
	}


	/**
	 * Compare local data access object with the ConnId object.
	 *
	 * @param co ConnId object
	 * @param o data access object
	 */
	private void check(final ConnectorObject co, final Generic o) {
		Assert.assertNotNull(co, "connector object");
		Assert.assertNotNull(co.getUid().getUidValue(), "operational __UID__");
		Assert.assertNotNull(co.getUid().getUidValue(), "operational __UID__ value");
		// compare operational UID only when explicitly specified
		if (o.getUid() != null) {
			Assert.assertEquals(co.getUid().getUidValue(), o.getUid(), "operational __UID__ value");
		}
		Assert.assertEquals(co.getName().getNameValue(), o.getName(), "operational __NAME__");
		for (String key : o.getData().keySet()) {
			// FIXME: check this, timemodified attribute is changing
			if (Cohort.ATTR_TIMEMODIFIED.equals(key)) {
				continue;
			}
			boolean multi = Collection.class.isAssignableFrom(o.ATTRS.get(key));
			final Attribute a = co.getAttributeByName(key);
			if (key.equals(o.getNameAttr())) {
				Assert.assertNull(a, "__NAME__ is not there as " + key);
			} else {
				Assert.assertNotNull(a, key + " attribute");
				Assert.assertNotNull(a.getValue(), key + " attribute value");
				if (!multi) {
					Assert.assertEquals(a.getValue().size(), 1, key + " attribute value cardinality for single-value attribute");
				}
				final Object v1 = multi ? a.getValue() : a.getValue().get(0);
				final Object v2 = o.getData().get(key);
				if (v1 instanceof GuardedString) {
					checkPwd(key, a.getValue(), (GuardedString) v2);
				} else {
					Assert.assertEquals(v1, v2, key + " attribute");
				}
			}
		}
	}


	/**
	 * Return connector for testing.
	 *
	 * @param environment testing environment
	 * @return connector
	 */
	protected ConnectorFacade getFacade(final String environment) {
		if (connectorFacade == null) {
			synchronized (this) {
				if (connectorFacade == null) {
					connectorFacade = createConnectorFacade(MoodleConnector.class, environment);
				}
			}
		}
		return connectorFacade;
	}


	/**
	 * Create connector facade for testing.
	 *
	 * @param clazz connector class
	 * @param environment testing environment
	 * @return connector for testing
	 */
	public ConnectorFacade createConnectorFacade(final Class<? extends Connector> clazz, final String environment) {
		final APIConfiguration impl =
				TestHelpers.createTestConfiguration(clazz, PROPERTIES, "configuration");
		impl.setProducerBufferSize(0);
		impl.getResultsHandlerConfiguration().setEnableAttributesToGetSearchResultsHandler(false);
		impl.getResultsHandlerConfiguration().setEnableCaseInsensitiveFilter(false);
		impl.getResultsHandlerConfiguration().setEnableFilteredResultsHandler(false);
		impl.getResultsHandlerConfiguration().setEnableNormalizingResultsHandler(false);

		//impl.setTimeout(CreateApiOp.class, 25000);
		//impl.setTimeout(UpdateApiOp.class, 25000);
		//impl.setTimeout(DeleteApiOp.class, 25000);

		return ConnectorFacadeFactory.getInstance().newInstance(impl);
	}

}
