package cz.zcu.connectors.moodle;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.identityconnectors.common.logging.Log;
import org.identityconnectors.common.security.GuardedString;
import org.identityconnectors.common.security.SecurityUtil;
import org.identityconnectors.framework.api.APIConfiguration;
import org.identityconnectors.framework.api.ConnectorFacade;
import org.identityconnectors.framework.api.ConnectorFacadeFactory;
import org.identityconnectors.framework.common.exceptions.AlreadyExistsException;
import org.identityconnectors.framework.common.exceptions.UnknownUidException;
import org.identityconnectors.framework.common.objects.Attribute;
import org.identityconnectors.framework.common.objects.AttributeBuilder;
import org.identityconnectors.framework.common.objects.AttributeDelta;
import org.identityconnectors.framework.common.objects.AttributeDeltaBuilder;
import org.identityconnectors.framework.common.objects.AttributeInfo;
import org.identityconnectors.framework.common.objects.ConnectorObject;
import org.identityconnectors.framework.common.objects.Name;
import org.identityconnectors.framework.common.objects.ObjectClass;
import org.identityconnectors.framework.common.objects.ObjectClassInfo;
import org.identityconnectors.framework.common.objects.OperationOptionsBuilder;
import org.identityconnectors.framework.common.objects.OperationalAttributes;
import org.identityconnectors.framework.common.objects.Schema;
import org.identityconnectors.framework.common.objects.Uid;
import org.identityconnectors.framework.common.objects.filter.Filter;
import org.identityconnectors.framework.common.objects.filter.FilterBuilder;
import org.identityconnectors.framework.spi.Connector;
import org.identityconnectors.test.common.PropertyBag;
import org.identityconnectors.test.common.TestHelpers;
import org.identityconnectors.test.common.ToListResultsHandler;
import org.testng.Assert;
import org.testng.TestNGException;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import cz.zcu.connectors.Db;
import cz.zcu.connectors.dao.Generic;
import cz.zcu.connectors.moodle.dao.User;


public class MoodleConnectorTests {
	private static final Log l = Log.getLog(MoodleConnectorTests.class);
	private static final PropertyBag PROPERTIES = TestHelpers.getProperties(MoodleConnector.class, null);
	private volatile ConnectorFacade connectorFacade;
	private User user, userWrite, userWrite2, userUnknown;
	private Db db;
	private boolean purge = false;


	@BeforeClass
	public void setUp() {
		MoodleConfiguration config = new MoodleConfiguration();
		Map<String, Object> configData = new HashMap<>();
		for (String k : Arrays.asList("jdbcDriver", "jdbcUrl", "jdbcUser", "deleteMethod")) {
			try {
				Object v;
				switch(k) {
				default:
					v = PROPERTIES.getProperty("configuration." + k, String.class);
				}
				if (v != null)
					configData.put(k, v);
			} catch (IllegalArgumentException iae) {
				// OK
			}
		}
		TestHelpers.fillConfiguration(config, configData);
		for (String k : Arrays.asList("jdbcPassword")) {
			try {
				GuardedString pwd;
				switch(k) {
				case "jdbcPassword":
					pwd = PROPERTIES.getProperty("configuration." + k, GuardedString.class);
					config.setJdbcPassword(pwd);
					break;
				default:
				}
			} catch (IllegalArgumentException iae) {
				// OK
			}
		}
		db = new Db("moodle", config);
		purge = MoodleConfiguration.DELETE_METHOD_PURGE.equals(config.getDeleteMethod());

		user = loadUser("user");
		userUnknown = loadUser("userUnknown");
		userWrite = loadUser("userWrite");
		userWrite2 = loadUser("userWrite2");

		delete(getFacade(null), userWrite);
		delete(getFacade(null), userWrite2);
	}


	@AfterClass
	public void tearDown() {
		if (connectorFacade != null)
			connectorFacade.dispose();
		db.close();
	}


	@BeforeMethod
	public void setUpEach() {
	}


	private Uid create(ConnectorFacade facade, Generic o) {
		delete(facade, o);
		if (o instanceof User)
			return facade.create(User.OBJECT_CLASS, o.getAttributes(), null);
		else
			throw new TestNGException("Unknown DAO type " + o);
	}


	private void delete(ConnectorFacade facade, Generic o) {
		PreparedStatement st;
		try {
			if (o instanceof User) {
				l.info("delete: {1}, uid {2}", o.getClass().getName(), o.getUid());
				final String sqlUser = "DELETE FROM mdl_user WHERE id = ?";
				final String sqlAttr = "DELETE FROM mdl_user_attribute WHERE user_id = ?";

				st = db.create(sqlAttr);
				st.setString(1, o.getUid());
				l.info("SQL: {0} [{1}]", sqlAttr, o.getUid());
				st.execute();
				l.info(" => deleted {0} rows", st.getUpdateCount());

				st = db.create(sqlUser);
				st.setString(1, o.getUid());
				l.info("SQL: {0} [{1}]", sqlUser, o.getUid());
				st.execute();
				l.info(" => deleted {0} rows", st.getUpdateCount());
			} else
				throw new TestNGException("Unknown DAO type " + o);
		} catch (SQLException sex) {
			throw new TestNGException("Error deleting", sex);
		}
	}


	private void load(Generic o, String name, Map<String,Class<?>> ATTRS) {
		String prefix = name == null || "".equals(name) ? "" : name + ".";

		for (String key : ATTRS.keySet()) {
			Class<?> type = ATTRS.get(key);
			Object value = PROPERTIES.getProperty(prefix + key, type, null);
			if (value != null) {
				o.getData().put(key, value);
			}
		}
	}


	private void loadUser(User user, String name) {
		load(user, name, User.ATTRS);
		String uid = user.get(User.ATTR_ID).toString();
		user.setUid(uid);
		user.setName(uid);
		user.remove(User.ATTR_ID);
		if (user.contains(User.ATTR_PASSWORD)) {
			user.setPassword(user.get(User.ATTR_PASSWORD));
			user.remove(User.ATTR_PASSWORD);
		}
	}


	private User loadUser(String name) {
		User user = new User();
		loadUser(user, name);
		return user;
	}


	private void getTest(ObjectClass objectClass, Generic o) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();

		Uid uid = new Uid(o.getUid());

		builder.setAttributesToGet(Name.NAME);
		ConnectorObject co = facade.getObject(objectClass, uid, builder.build());
		check(co, o);
	}


	@Test
	public void getUserTest() {
		getTest(User.OBJECT_CLASS, user);
	}


	private void getUnknownTest(ObjectClass objectClass, Generic o) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();
		ConnectorObject co;

		builder.setAttributesToGet(Name.NAME);
		co = facade.getObject(objectClass, new Uid(o.getUid()), builder.build());
		Assert.assertNull(co);
	}


	@Test
	public void getUnknownUserTest() {
		getUnknownTest(User.OBJECT_CLASS, userUnknown);
	}


	private void searchExactTest(ObjectClass objectClass, Generic o, Filter[] FILTERS) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();
		ToListResultsHandler handler;
		ConnectorObject co;

		for (Filter f : FILTERS) {
			handler = new ToListResultsHandler();
			facade.search(objectClass, f, handler, builder.build());
			Assert.assertEquals(handler.getObjects().size(), 1, "exactly one user found");
			co = handler.getObjects().get(0);
			check(co, o);
		}
	}


	@Test
	public void searchExactUserTest() {
		Filter[] filters = {
			FilterBuilder.equalTo(new Uid(user.getUid())),
			FilterBuilder.equalTo(new Name(user.getName())),
			FilterBuilder.equalTo(AttributeBuilder.build(User.ATTR_ID, user.getUid())),
			FilterBuilder.equalTo(AttributeBuilder.build(User.ATTR_USERNAME, (String)user.get(User.ATTR_USERNAME))),
			FilterBuilder.equalTo(AttributeBuilder.build(User.ATTR_EMAIL, (String)user.get(User.ATTR_EMAIL))),
		};
		searchExactTest(User.OBJECT_CLASS, user, filters);
	}


	private void searchAllTest(ObjectClass objectClass, Generic o) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();
		builder.setPageSize(10);

		final ToListResultsHandler handler = new ToListResultsHandler();
		facade.search(objectClass, null, handler, builder.build());
		Assert.assertTrue(handler.getObjects().size() >= 2, "objects found");
		Assert.assertTrue(handler.getObjects().stream().anyMatch(it -> o.getUid().equals(it.getUid().getUidValue
())), "test object found");
	}


	@Test
	public void searchAllUserTest() {
		searchAllTest(User.OBJECT_CLASS, user);
	}


	private void createTest(ObjectClass objectClass, Generic o) {
		final ConnectorFacade facade = getFacade(null);
		ConnectorObject co;

		Uid uid = create(facade, o);
		Assert.assertNotNull(uid, "object created and __UID__ attribute not null");
		Assert.assertNotNull(uid.getUidValue(), "__UID__ attribute not empty");
		Assert.assertEquals(uid.getUidValue(), o.getUid(), "__UID__ attribute value");

		co = facade.getObject(objectClass, uid, null);
		check(co, o);
	}


	@Test
	public void createUserTest() {
		createTest(User.OBJECT_CLASS, userWrite);
	}


	private void deleteTest(ObjectClass objectClass, Generic o) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();
		ConnectorObject co;

		create(facade, o);

		facade.delete(objectClass, new Uid(o.getUid()), builder.build());
		co = facade.getObject(objectClass, new Uid(o.getUid()), builder.build());
		if (purge)
			Assert.assertNull(co, "object really deleted");
		else {
			Assert.assertNotNull(co, "object not really deleted");
			Attribute a = co.getAttributeByName(User.ATTR_USERNAME);
			Assert.assertNotNull(a, "attribute " + User.ATTR_USERNAME);
			List<Object> lo = a.getValue();
			Assert.assertNotNull(lo, "attribute value " + User.ATTR_USERNAME);
			Assert.assertFalse(lo.isEmpty(), "attribute value " + User.ATTR_USERNAME + " empty");
			Assert.assertTrue(((String)lo.get(0)).contains("."), User.ATTR_USERNAME + " updated for recycling");
		}
	}


	@Test
	public void deleteUserTest() {
		deleteTest(User.OBJECT_CLASS, userWrite);
	}


	private void deleteUnknownTest(ObjectClass objectClass, Generic o) {
		final ConnectorFacade facade = getFacade(null);

		facade.delete(objectClass, new Uid(o.getUid()), null);
	}


	@Test(expectedExceptions = { UnknownUidException.class })
	public void deleteUnknownUserTest() {
		deleteUnknownTest(User.OBJECT_CLASS, userUnknown);
	}


	/*
	 * Check related changes after changing UID attribute
	 */
	private void checkChanges(Set<AttributeDelta> changes, String newUid, String... changedAttrs) {
		Assert.assertNotNull(changes, "changes");
		Assert.assertEquals(changes.size(), changedAttrs.length, "changes has " + changedAttrs.length + " elements");
		Set<String> found = changes.stream()
			.map(AttributeDelta::getName)
			.collect(Collectors.toSet());
		Set<String> expected = new HashSet<>(Arrays.asList(changedAttrs));
		Assert.assertEquals(found, expected, "changes");
	}


	private void updateTest(ObjectClass objectClass, Generic o, Generic o2, boolean withCreate, boolean withRename) {
		final ConnectorFacade facade = getFacade(null);
		ConnectorObject co;
		Set<AttributeDelta> modifications;

		delete(facade, o);
		delete(facade, o2);
		if (withCreate)
			create(facade, o);
		String uidOrig = o.getUid();
		Generic oNew = o2.clone();
		if (!withRename) {
			oNew.setUid(uidOrig);
			oNew.setName(uidOrig);
		}

		modifications = oNew.getDelta();
		facade.updateDelta(objectClass, new Uid(uidOrig), modifications, null);
		co = facade.getObject(objectClass, new Uid(oNew.getUid()), null);
		check(co, oNew);
	}


	@Test
	public void updateUserTest() {
		updateTest(User.OBJECT_CLASS, userWrite, userWrite2, true, false);
	}


	@Test(expectedExceptions = { UnknownUidException.class })
	public void updateUnknownUserTest() {
		updateTest(User.OBJECT_CLASS, userWrite, userWrite2, false, false);
	}


	@Test
	public void updateWithRenameUserTest() {
		updateTest(User.OBJECT_CLASS, userWrite, userWrite2, true, true);
	}


	private void renameTest(ObjectClass objectClass, Generic o, Generic o2) {
		final ConnectorFacade facade = getFacade(null);
		final OperationOptionsBuilder builder = new OperationOptionsBuilder();

		delete(facade, o);
		delete(facade, o2);
		create(facade, o);

		Uid uid = new Uid(o.getUid());
		String name = o.getName();
		String oldName = o.getName();
		for (String newName : Arrays.asList(name, o2.getName())) {
			l.info("Renaming {0} -> {1}", name, newName);
			Set<AttributeDelta> modifications = new HashSet<>();

			modifications.add(AttributeDeltaBuilder.build(Name.NAME, newName));
			Set<AttributeDelta> changes = facade.updateDelta(objectClass, uid, modifications, builder.build());
			// changing NAME will change UID too as NAME and UID correspond to identical database column
			if (oldName.equals(newName))
				Assert.assertNull(changes, "changes must be null on identical rename");
			else
				checkChanges(changes, newName, Uid.NAME);

			ToListResultsHandler handler = new ToListResultsHandler();
			facade.search(objectClass, FilterBuilder.equalTo(new Name(newName)), handler, null);
			Assert.assertNotNull(handler.getObjects(), "any objects found by new name");
			Assert.assertEquals(handler.getObjects().size(), 1, "exactly one user found by new name");
			ConnectorObject co = handler.getObjects().get(0);
			Assert.assertNotNull(co, "user found by the new name");
			Assert.assertEquals(co.getName().getNameValue(), newName, "NAME is the new name");

			Generic oNew = o.clone();
			oNew.setUid(newName);
			oNew.setName(newName);
			check(co, oNew);

			oldName = newName;
		}
	}


	@Test
	public void renameUserTest() {
		renameTest(User.OBJECT_CLASS, userWrite, userWrite2);
	}


	private void renameExistsTest(ObjectClass objectClass, Generic o, Generic o2) {
		final ConnectorFacade facade = getFacade(null);

		delete(facade, o);
		delete(facade, o2);
		create(facade, o);
		create(facade, o2);

		Set<AttributeDelta> modifications = new HashSet<>();
		modifications.add(AttributeDeltaBuilder.build(Name.NAME, o2.getName()));
		facade.updateDelta(objectClass, new Uid(o.getUid()), modifications, null);
	}


	@Test(expectedExceptions = { AlreadyExistsException.class })
	public void renameExistsUserTest() {
		renameExistsTest(User.OBJECT_CLASS, userWrite, userWrite2);
	}


	private void updateResetTest(ObjectClass objectClass, Generic o, Generic o2, boolean emptyString, String... ATTRS) {
		final ConnectorFacade facade = getFacade(null);
		ConnectorObject co;
		Set<AttributeDelta> modifications = new HashSet<>();

		delete(facade, o);
		delete(facade, o2);
		create(facade, o);

		List<Object> nullValue = new ArrayList<>();
		if (emptyString)
			nullValue.add("");
		else
			nullValue.add(null);
		for (String attr : ATTRS)
			modifications.add(AttributeDeltaBuilder.build(attr, nullValue));
		Set<AttributeDelta> changes = facade.updateDelta(objectClass, new Uid(o.getUid()), modifications, null);
		if (emptyString)
			checkChanges(changes, o.getUid(), ATTRS);
		else
			Assert.assertNull(changes, "changes");

		co = facade.getObject(objectClass, new Uid(o.getUid()), null);
		for (String attr : ATTRS) {
			Attribute a = co.getAttributeByName(attr);
			Assert.assertNull(a, attr + " attribute");
		}
	}


	@Test
	public void updateResetUserTest() {
		String[] NULLABLE_ATTRS = {
				User.ATTR_USERNAME,
				User.ATTR_FIRSTNAME,
				User.ATTR_LASTNAME,
				User.ATTR_EMAIL,
				User.ATTR_INSTITUTION,
				User.ATTR_PASSWORD,
				User.ATTR_LANG,
				User.ATTR_PERSONAL_NUMBER,
		};
		updateResetTest(User.OBJECT_CLASS, userWrite, userWrite2, false, NULLABLE_ATTRS);
	}


	@Test
	public void updateResetByEmptyUserTest() {
		// User.ATTR_PERSONAL_NUMBER is not nullable by empty string, because empty strings are allowed in user attributes table
		String[] NULLABLE_ATTRS = {
				User.ATTR_USERNAME,
				User.ATTR_FIRSTNAME,
				User.ATTR_LASTNAME,
				User.ATTR_EMAIL,
				User.ATTR_INSTITUTION,
				User.ATTR_LANG,
		};
		updateResetTest(User.OBJECT_CLASS, userWrite, userWrite2, true, NULLABLE_ATTRS);
	}


	@Test
	public void schemaTest() {
		final ConnectorFacade facade = getFacade(null);
		ObjectClassInfo info;
		Set<String> expected;
		String[] expectedArray = {
			Name.NAME,
			Uid.NAME,
			OperationalAttributes.PASSWORD_NAME,
			// operational: User.ATTR_ID,
			User.ATTR_AUTH,
			User.ATTR_MNETHOSTID,
			User.ATTR_USERNAME,
			User.ATTR_FIRSTNAME,
			User.ATTR_LASTNAME,
			User.ATTR_EMAIL,
			User.ATTR_INSTITUTION,
			User.ATTR_CONFIRMED,
			User.ATTR_DELETED,
			User.ATTR_LANG,
		};

		Schema schema = facade.schema();

		expected = new HashSet<>(Arrays.asList(expectedArray));
		info = schema.findObjectClassInfo(User.OBJECT_CLASS_NAME);
		for (AttributeInfo a : info.getAttributeInfo()) {
			l.info("attribute {0}", a.getName());
			switch (a.getName()) {
			case "__NAME__":
				Assert.assertEquals(User.ATTR_ID, a.getNativeName(), "__NAME__ is " + User.ATTR_ID);
				break;
			case "__UID__":
				Assert.assertEquals(User.ATTR_ID, a.getNativeName(), "__UID__ is " + User.ATTR_ID);
				break;
			}
			expected.remove(a.getName());
			// forbidden in midPoint
			Assert.assertNotEquals(a.getName(), "id");
		}
		l.info("remaining missed attributes: {0}", expected);
		Assert.assertTrue(expected.isEmpty(), "no missing attributes from user");
	}


	@Test
	public void testTest() {
		final ConnectorFacade facade = getFacade(null);
		facade.test();
	}


	@Test
	public void validateTest() {
		final ConnectorFacade facade = getFacade(null);
		facade.validate();
	}


	private void checkPwd(final String attrName, List<Object>values, GuardedString expectedPassword) {
		final String MSG = attrName + " attribute value";

		if (values == null || values.isEmpty()) {
			Assert.assertNull(expectedPassword, MSG);
			return;
		}
		GuardedString foundPassword = (GuardedString)values.get(0);
		if (foundPassword == null || expectedPassword == null)
			Assert.assertEquals(foundPassword, expectedPassword, MSG);
		else
			Assert.assertEquals(SecurityUtil.decrypt(foundPassword), SecurityUtil.decrypt(expectedPassword), MSG);
	}


	private void check(ConnectorObject co, Generic o) {
		Assert.assertNotNull(co, "connector object");
		Assert.assertEquals(co.getUid().getUidValue(), o.getUid(), "__UID__");
		Assert.assertEquals(co.getName().getNameValue(), o.getName(), "__NAME__");
		checkPwd(OperationalAttributes.PASSWORD_NAME, co.getAttributeByName(OperationalAttributes.PASSWORD_NAME).getValue(), o.getPassword());
		for (String key : o.getData().keySet()) {
			Attribute a = co.getAttributeByName(key);
			Assert.assertNotNull(a, key + " attribute");
			Assert.assertNotNull(a.getValue(), key + " attribute value");
			Assert.assertEquals(a.getValue().size(), 1, key + " attribute cardinality");
			Object v1 = a.getValue().get(0);
			Object v2 = o.getData().get(key);
			if (v1 instanceof GuardedString) {
				checkPwd(key, a.getValue(), (GuardedString)v2);
			} else {
				Assert.assertEquals(v1, v2, key + " attribute");
			}
		}
	}


	protected ConnectorFacade getFacade(String environment) {
		if (connectorFacade == null) {
			synchronized (this) {
				if (connectorFacade == null) {
					connectorFacade = createConnectorFacade(MoodleConnector.class, environment);
				}
			}
		}
		return connectorFacade;
	}


	public ConnectorFacade createConnectorFacade(Class<? extends Connector> clazz, String environment) {
		APIConfiguration impl =
				TestHelpers.createTestConfiguration(clazz, PROPERTIES, "configuration");
		impl.setProducerBufferSize(0);
		impl.getResultsHandlerConfiguration().setEnableAttributesToGetSearchResultsHandler(false);
		impl.getResultsHandlerConfiguration().setEnableCaseInsensitiveFilter(false);
		impl.getResultsHandlerConfiguration().setEnableFilteredResultsHandler(false);
		impl.getResultsHandlerConfiguration().setEnableNormalizingResultsHandler(false);

		//impl.setTimeout(CreateApiOp.class, 25000);
		//impl.setTimeout(UpdateApiOp.class, 25000);
		//impl.setTimeout(DeleteApiOp.class, 25000);

		return ConnectorFacadeFactory.getInstance().newInstance(impl);
	}

}
